# distutils: libraries = STD_LIBS
"""

This file contains definitions of the various primitives that can be used
by the Cython ray-tracer for unstructured mesh rendering. To define a new
primitive type, you need to define a struct that represents it. You also
need to provide three functions:

1. A function that computes the intersection between a given ray and a given primitive.
2. A function that computes the centroid of the primitive type.
3. A function that computes the axis-aligned bounding box of a given primitive.

"""

cimport cython

import numpy as np

cimport numpy as np
from libc.math cimport fabs

from yt.utilities.lib.vec3_ops cimport L2_norm, cross, distance, dot, subtract


cdef np.float64_t DETERMINANT_EPS = 1.0e-10
cdef np.float64_t INF = np.inf

cdef extern from "platform_dep.h" nogil:
    double fmax(double x, double y)
    double fmin(double x, double y)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef np.int64_t ray_bbox_intersect(Ray* ray, const BBox bbox) nogil:
    '''

    This returns an integer flag that indicates whether a ray and a bounding
    box intersect. It does not modify either either the ray or the box.

    '''

    # https://tavianator.com/fast-branchless-raybounding-box-intersections/

    cdef np.float64_t tmin = -INF
    cdef np.float64_t tmax =  INF

    cdef np.int64_t i
    cdef np.float64_t t1, t2
    for i in range(3):
        t1 = (bbox.left_edge[i]  - ray.origin[i])*ray.inv_dir[i]
        t2 = (bbox.right_edge[i] - ray.origin[i])*ray.inv_dir[i]
        tmin = fmax(tmin, fmin(t1, t2))
        tmax = fmin(tmax, fmax(t1, t2))

    return tmax >= fmax(tmin, 0.0)

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef np.int64_t ray_triangle_intersect(const void* primitives,
                                       const np.int64_t item,
                                       Ray* ray) nogil:
    '''

    This returns an integer flag that indicates whether a triangle is the
    closest hit for the ray so far. If it is, the ray is updated to store the
    current triangle index and the distance to the first hit. The triangle used
    is the one indexed by "item" in the array of primitives.


    '''

    # https://en.wikipedia.org/wiki/M%C3%B6ller%E2%80%93Trumbore_intersection_algorithm

    cdef Triangle tri = (<Triangle*> primitives)[item]

    # edge vectors
    cdef np.float64_t e1[3]
    cdef np.float64_t e2[3]
    subtract(tri.p1, tri.p0, e1)
    subtract(tri.p2, tri.p0, e2)

    cdef np.float64_t P[3]
    cross(ray.direction, e2, P)

    cdef np.float64_t det, inv_det
    det = dot(e1, P)
    if(det > -DETERMINANT_EPS and det < DETERMINANT_EPS):
        return False
    inv_det = 1.0 / det

    cdef np.float64_t T[3]
    subtract(ray.origin, tri.p0, T)

    cdef np.float64_t u = dot(T, P) * inv_det
    if(u < 0.0 or u > 1.0):
        return False

    cdef np.float64_t Q[3]
    cross(T, e1, Q)

    cdef np.float64_t v = dot(ray.direction, Q) * inv_det
    if(v < 0.0 or u + v  > 1.0):
        return False

    cdef np.float64_t t = dot(e2, Q) * inv_det

    if(t > DETERMINANT_EPS and t < ray.t_far):
        ray.t_far = t
        ray.elem_id = tri.elem_id
        return True

    return False


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void triangle_centroid(const void *primitives,
                            const np.int64_t item,
                            np.float64_t[3] centroid) nogil:
    '''

    This computes the centroid of the input triangle. The triangle used
    is the one indexed by "item" in the array of primitives. The result
    will be stored in the numpy array passed in as "centroid".

    '''

    cdef Triangle tri = (<Triangle*> primitives)[item]
    cdef np.int64_t i
    for i in range(3):
        centroid[i] = (tri.p0[i] + tri.p1[i] + tri.p2[i]) / 3.0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void triangle_bbox(const void *primitives,
                        const np.int64_t item,
                        BBox* bbox) nogil:
    '''

    This computes the bounding box of the input triangle. The triangle used
    is the one indexed by "item" in the array of primitives. The result
    will be stored in the input BBox.

    '''

    cdef Triangle tri = (<Triangle*> primitives)[item]
    cdef np.int64_t i
    for i in range(3):
        bbox.left_edge[i] = fmin(fmin(tri.p0[i], tri.p1[i]), tri.p2[i])
        bbox.right_edge[i] = fmax(fmax(tri.p0[i], tri.p1[i]), tri.p2[i])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void patchSurfaceFunc(const cython.floating[8][3] verts,
                           const cython.floating u,
                           const cython.floating v,
                           cython.floating[3] S) nogil:
    '''

    This function is a parametric representation of the surface of a bi-quadratic
    patch. The inputs are the eight nodes that define a face of a 20-node hex element,
    and two parameters u and v that vary from -1 to 1 and tell you where you are on
    the surface of the patch. The output is the array 'S' that stores the physical
    (x, y, z) position of the corresponding point on the patch. This function is needed
    to compute the intersection of rays and bi-quadratic patches.

    '''
    cdef int i
    for i in range(3):
        S[i] = 0.25*(1.0 - u)*(1.0 - v)*(-u - v - 1)*verts[0][i] + \
               0.25*(1.0 + u)*(1.0 - v)*( u - v - 1)*verts[1][i] + \
               0.25*(1.0 + u)*(1.0 + v)*( u + v - 1)*verts[2][i] + \
               0.25*(1.0 - u)*(1.0 + v)*(-u + v - 1)*verts[3][i] + \
               0.5*(1 - u)*(1 - v*v)*verts[4][i] + \
               0.5*(1 - u*u)*(1 - v)*verts[5][i] + \
               0.5*(1 + u)*(1 - v*v)*verts[6][i] + \
               0.5*(1 - u*u)*(1 + v)*verts[7][i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void patchSurfaceDerivU(const cython.floating[8][3] verts,
                             const cython.floating u,
                             const cython.floating v,
                             cython.floating[3] Su) nogil:
    '''

    This function computes the derivative of the S(u, v) function w.r.t u.

    '''
    cdef int i
    for i in range(3):
        Su[i] = (-0.25*(v - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \
                (-0.25*(v - 1.0)*(u - v - 1) - 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
                ( 0.25*(v + 1.0)*(u + v - 1) + 0.25*(u + 1.0)*(v + 1.0))*verts[2][i] + \
                ( 0.25*(v + 1.0)*(u - v + 1) + 0.25*(u - 1.0)*(v + 1.0))*verts[3][i] + \
                0.5*(v*v - 1.0)*verts[4][i] + u*(v - 1.0)*verts[5][i] - \
                0.5*(v*v - 1.0)*verts[6][i] - u*(v + 1.0)*verts[7][i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void patchSurfaceDerivV(const cython.floating[8][3] verts,
                             const cython.floating u,
                             const cython.floating v,
                             cython.floating[3] Sv) nogil:
    '''

    This function computes the derivative of the S(u, v) function w.r.t v.

    '''

    cdef int i
    for i in range(3):
        Sv[i] = (-0.25*(u - 1.0)*(u + v + 1) - 0.25*(u - 1.0)*(v - 1.0))*verts[0][i] + \
                (-0.25*(u + 1.0)*(u - v - 1) + 0.25*(u + 1.0)*(v - 1.0))*verts[1][i] + \
                ( 0.25*(u + 1.0)*(u + v - 1) + 0.25*(u + 1.0)*(v + 1.0))*verts[2][i] + \
                ( 0.25*(u - 1.0)*(u - v + 1) - 0.25*(u - 1.0)*(v + 1.0))*verts[3][i] + \
                0.5*(u*u - 1.0)*verts[5][i] + v*(u - 1.0)*verts[4][i] - \
                0.5*(u*u - 1.0)*verts[7][i] - v*(u + 1.0)*verts[6][i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef RayHitData compute_patch_hit(cython.floating[8][3] verts,
                                  cython.floating[3] ray_origin,
                                  cython.floating[3] ray_direction) nogil:
    """

    This function iteratively computes whether the bi-quadratic patch defined by the
    eight input nodes intersects with the given ray. Either way, information about
    the potential hit is stored in the returned RayHitData.

    """
    # first we compute the two planes that define the ray.
    cdef cython.floating[3] n, N1, N2
    cdef cython.floating A = dot(ray_direction, ray_direction)
    for i in range(3):
        n[i] = ray_direction[i] / A

    if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
        N1[0] = n[1]
        N1[1] =-n[0]
        N1[2] = 0.0
    else:
        N1[0] = 0.0
        N1[1] = n[2]
        N1[2] =-n[1]
    cross(N1, n, N2)

    cdef cython.floating d1 = -dot(N1, ray_origin)
    cdef cython.floating d2 = -dot(N2, ray_origin)

    # the initial guess is set to zero
    cdef cython.floating u = 0.0
    cdef cython.floating v = 0.0
    cdef cython.floating[3] S
    patchSurfaceFunc(verts, u, v, S)
    cdef cython.floating fu = dot(N1, S) + d1
    cdef cython.floating fv = dot(N2, S) + d2
    cdef cython.floating err = fmax(fabs(fu), fabs(fv))

    # begin Newton iteration
    cdef cython.floating tol = 1.0e-5
    cdef int iterations = 0
    cdef int max_iter = 10
    cdef cython.floating[3] Su
    cdef cython.floating[3] Sv
    cdef cython.floating J11, J12, J21, J22, det
    while ((err > tol) and (iterations < max_iter)):
        # compute the Jacobian
        patchSurfaceDerivU(verts, u, v, Su)
        patchSurfaceDerivV(verts, u, v, Sv)
        J11 = dot(N1, Su)
        J12 = dot(N1, Sv)
        J21 = dot(N2, Su)
        J22 = dot(N2, Sv)
        det = (J11*J22 - J12*J21)

        # update the u, v values
        u -= ( J22*fu - J12*fv) / det
        v -= (-J21*fu + J11*fv) / det

        patchSurfaceFunc(verts, u, v, S)
        fu = dot(N1, S) + d1
        fv = dot(N2, S) + d2

        err = fmax(fabs(fu), fabs(fv))
        iterations += 1

    # t is the distance along the ray to this hit
    cdef cython.floating t = distance(S, ray_origin) / L2_norm(ray_direction)

    # return hit data
    cdef RayHitData hd
    hd.u = u
    hd.v = v
    hd.t = t
    hd.converged = (iterations < max_iter)
    return hd


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef np.int64_t ray_patch_intersect(const void* primitives,
                                    const np.int64_t item,
                                    Ray* ray) nogil:
    '''

    This returns an integer flag that indicates whether the given patch is the
    closest hit for the ray so far. If it is, the ray is updated to store the
    current primitive index and the distance to the first hit. The patch used
    is the one indexed by "item" in the array of primitives.


    '''
    cdef Patch patch = (<Patch*> primitives)[item]

    cdef RayHitData hd = compute_patch_hit(patch.v, ray.origin, ray.direction)

    # only count this is it's the closest hit
    if (hd.t < ray.t_near or hd.t > ray.t_far):
        return False

    if (fabs(hd.u) <= 1.0 and fabs(hd.v) <= 1.0 and hd.converged):
        # we have a hit, so update ray information
        ray.t_far = hd.t
        ray.elem_id = patch.elem_id
        return True

    return False


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void patch_centroid(const void *primitives,
                         const np.int64_t item,
                         np.float64_t[3] centroid) nogil:
    '''

    This computes the centroid of the input patch. The patch used
    is the one indexed by "item" in the array of primitives. The result
    will be stored in the numpy array passed in as "centroid".

    '''

    cdef np.int64_t i, j
    cdef Patch patch = (<Patch*> primitives)[item]

    for j in range(3):
        centroid[j] = 0.0

    for i in range(8):
        for j in range(3):
            centroid[j] += patch.v[i][j]

    for j in range(3):
        centroid[j] /= 8.0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void patch_bbox(const void *primitives,
                    const np.int64_t item,
                     BBox* bbox) nogil:

    '''

    This computes the bounding box of the input patch. The patch used
    is the one indexed by "item" in the array of primitives. The result
    will be stored in the input BBox.

    '''

    cdef np.int64_t i, j
    cdef Patch patch = (<Patch*> primitives)[item]

    for j in range(3):
        bbox.left_edge[j] = patch.v[0][j]
        bbox.right_edge[j] = patch.v[0][j]

    for i in range(1, 8):
        for j in range(3):
            bbox.left_edge[j] = fmin(bbox.left_edge[j], patch.v[i][j])
            bbox.right_edge[j] = fmax(bbox.right_edge[j], patch.v[i][j])


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void tet_patchSurfaceFunc(const cython.floating[6][3] verts,
                           const cython.floating u,
                           const cython.floating v,
                           cython.floating[3] S) nogil:

    cdef int i
    # Computes for canonical triangle coordinates
    for i in range(3):
        S[i] = (1.0 - 3.0*u + 2.0*u*u - 3.0*v + 2.0*v*v + 4.0*u*v)*verts[0][i] + \
             (-u + 2.0*u*u)*verts[1][i] + \
             (-v + 2.0*v*v)*verts[2][i] + \
             (4.0*u - 4.0*u*u - 4.0*u*v)*verts[3][i] + \
             (4.0*u*v)*verts[4][i] + \
             (4.0*v - 4.0*v*v - 4.0*u*v)*verts[5][i]

@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void tet_patchSurfaceDerivU(const cython.floating[6][3] verts,
                             const cython.floating u,
                             const cython.floating v,
                             cython.floating[3] Su) nogil:
    cdef int i
    # Computes for canonical triangle coordinates
    for i in range(3):
        Su[i] = (-3.0 + 4.0*u + 4.0*v)*verts[0][i] + \
             (-1.0 + 4.0*u)*verts[1][i] + \
             (4.0 - 8.0*u - 4.0*v)*verts[3][i] + \
             (4.0*v)*verts[4][i] + \
             (-4.0*v)*verts[5][i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void tet_patchSurfaceDerivV(const cython.floating[6][3] verts,
                             const cython.floating u,
                             const cython.floating v,
                             cython.floating[3] Sv) nogil:

    cdef int i
    # Computes for canonical triangle coordinates
    for i in range(3):
        Sv[i] = (-3.0 + 4.0*v + 4.0*u)*verts[0][i] + \
             (-1.0 + 4.0*v)*verts[2][i] + \
             (-4.0*u)*verts[3][i] + \
             (4.0*u)*verts[4][i] + \
             (4.0 - 8.0*v - 4.0*u)*verts[5][i]


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef RayHitData compute_tet_patch_hit(cython.floating[6][3] verts,
                                  cython.floating[3] ray_origin,
                                  cython.floating[3] ray_direction) nogil:

    # first we compute the two planes that define the ray.
    cdef cython.floating[3] n, N1, N2
    cdef cython.floating A = dot(ray_direction, ray_direction)
    for i in range(3):
        n[i] = ray_direction[i] / A

    if ((fabs(n[0]) > fabs(n[1])) and (fabs(n[0]) > fabs(n[2]))):
        N1[0] = n[1]
        N1[1] =-n[0]
        N1[2] = 0.0
    else:
        N1[0] = 0.0
        N1[1] = n[2]
        N1[2] =-n[1]
    cross(N1, n, N2)

    cdef cython.floating d1 = -dot(N1, ray_origin)
    cdef cython.floating d2 = -dot(N2, ray_origin)

    # the initial guess is set to zero
    cdef cython.floating u = 0.0
    cdef cython.floating v = 0.0
    cdef cython.floating[3] S
    tet_patchSurfaceFunc(verts, u, v, S)
    cdef cython.floating fu = dot(N1, S) + d1
    cdef cython.floating fv = dot(N2, S) + d2
    cdef cython.floating err = fmax(fabs(fu), fabs(fv))

    # begin Newton iteration
    cdef cython.floating tol = 1.0e-5
    cdef int iterations = 0
    cdef int max_iter = 10
    cdef cython.floating[3] Su
    cdef cython.floating[3] Sv
    cdef cython.floating J11, J12, J21, J22, det
    while ((err > tol) and (iterations < max_iter)):
        # compute the Jacobian
        tet_patchSurfaceDerivU(verts, u, v, Su)
        tet_patchSurfaceDerivV(verts, u, v, Sv)
        J11 = dot(N1, Su)
        J12 = dot(N1, Sv)
        J21 = dot(N2, Su)
        J22 = dot(N2, Sv)
        det = (J11*J22 - J12*J21)

        # update the u, v values
        u -= ( J22*fu - J12*fv) / det
        v -= (-J21*fu + J11*fv) / det

        tet_patchSurfaceFunc(verts, u, v, S)
        fu = dot(N1, S) + d1
        fv = dot(N2, S) + d2

        err = fmax(fabs(fu), fabs(fv))
        iterations += 1

    # t is the distance along the ray to this hit
    cdef cython.floating t = distance(S, ray_origin) / L2_norm(ray_direction)

    # return hit data
    cdef RayHitData hd
    hd.u = u
    hd.v = v
    hd.t = t
    hd.converged = (iterations < max_iter)
    return hd


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef np.int64_t ray_tet_patch_intersect(const void* primitives,
                                    const np.int64_t item,
                                    Ray* ray) nogil:

    cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]

    cdef RayHitData hd = compute_tet_patch_hit(tet_patch.v, ray.origin, ray.direction)

    # only count this is it's the closest hit
    if (hd.t < ray.t_near or hd.t > ray.t_far):
        return False

    if (0 <= hd.u and 0 <= hd.v and hd.u + hd.v <= 1.0 and hd.converged):
        # we have a hit, so update ray information
        ray.t_far = hd.t
        ray.elem_id = tet_patch.elem_id
        return True

    return False


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void tet_patch_centroid(const void *primitives,
                         const np.int64_t item,
                         np.float64_t[3] centroid) nogil:

    cdef np.int64_t i, j
    cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]

    for j in range(3):
        centroid[j] = 0.0

    for i in range(6):
        for j in range(3):
            centroid[j] += tet_patch.v[i][j]

    for j in range(3):
        centroid[j] /= 6.0


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
cdef void tet_patch_bbox(const void *primitives,
                    const np.int64_t item,
                     BBox* bbox) nogil:

    cdef np.int64_t i, j
    cdef TetPatch tet_patch = (<TetPatch*> primitives)[item]

    for j in range(3):
        bbox.left_edge[j] = tet_patch.v[0][j]
        bbox.right_edge[j] = tet_patch.v[0][j]

    for i in range(1, 6):
        for j in range(3):
            bbox.left_edge[j] = fmin(bbox.left_edge[j], tet_patch.v[i][j])
            bbox.right_edge[j] = fmax(bbox.right_edge[j], tet_patch.v[i][j])



import itertools
import logging
import os
import sys
import traceback
from functools import wraps
from io import StringIO
from typing import List

import numpy as np
from more_itertools import always_iterable

import yt.utilities.logger
from yt.config import ytcfg
from yt.data_objects.image_array import ImageArray
from yt.funcs import is_sequence
from yt.units.unit_registry import UnitRegistry  # type: ignore
from yt.units.yt_array import YTArray
from yt.utilities.exceptions import YTNoDataInObjectError
from yt.utilities.lib.quad_tree import QuadTree, merge_quadtrees
from yt.utilities.logger import ytLogger as mylog

# We default to *no* parallelism unless it gets turned on, in which case this
# will be changed.
MPI = None
parallel_capable = False

dtype_names = {
    "float32": "MPI.FLOAT",
    "float64": "MPI.DOUBLE",
    "int32": "MPI.INT",
    "int64": "MPI.LONG",
    "c": "MPI.CHAR",
}
op_names = {"sum": "MPI.SUM", "min": "MPI.MIN", "max": "MPI.MAX"}


class FilterAllMessages(logging.Filter):
    """
    This is a simple filter for logging.Logger's that won't let any
    messages pass.
    """

    def filter(self, record):
        return 0


# Set up translation table and import things


def traceback_writer_hook(file_suffix=""):
    def write_to_file(exc_type, exc, tb):
        sys.__excepthook__(exc_type, exc, tb)
        fn = f"yt_traceback{file_suffix}"
        with open(fn, "w") as fhandle:
            traceback.print_exception(exc_type, exc, tb, file=fhandle)
            print(f"Wrote traceback to {fn}")
        MPI.COMM_WORLD.Abort(1)

    return write_to_file


def default_mpi_excepthook(exception_type, exception_value, tb):
    traceback.print_tb(tb)
    mylog.error("%s: %s", exception_type.__name__, exception_value)
    comm = yt.communication_system.communicators[-1]
    if comm.size > 1:
        mylog.error("Error occurred on rank %d.", comm.rank)
    MPI.COMM_WORLD.Abort(1)


def enable_parallelism(suppress_logging: bool = False, communicator=None) -> bool:
    """
    This method is used inside a script to turn on MPI parallelism, via
    mpi4py.  More information about running yt in parallel can be found
    here: https://yt-project.org/docs/3.0/analyzing/parallel_computation.html

    Parameters
    ----------
    suppress_logging : bool
       If set to True, only rank 0 will log information after the initial
       setup of MPI.

    communicator : mpi4py.MPI.Comm
        The MPI communicator to use. This controls which processes yt can see.
        If not specified, will be set to COMM_WORLD.

    Returns
    -------
    parallel_capable: bool
        True if the call was successful. False otherwise.
    """
    global parallel_capable, MPI
    try:
        from mpi4py import MPI as _MPI
    except ImportError:
        mylog.error("Could not enable parallelism: mpi4py is not installed")
        return False
    MPI = _MPI
    exe_name = os.path.basename(sys.executable)

    # if no communicator specified, set to COMM_WORLD
    if communicator is None:
        communicator = MPI.COMM_WORLD

    parallel_capable = communicator.size > 1
    if not parallel_capable:
        mylog.error(
            "Could not enable parallelism: only one mpi process is running. "
            "To remedy this, launch the Python interpreter as\n"
            "  mpirun -n <X> python3 <yourscript>.py  # with X > 1 ",
        )
        return False

    mylog.info(
        "Global parallel computation enabled: %s / %s",
        communicator.rank,
        communicator.size,
    )
    communication_system.push(communicator)
    ytcfg["yt", "internals", "global_parallel_rank"] = communicator.rank
    ytcfg["yt", "internals", "global_parallel_size"] = communicator.size
    ytcfg["yt", "internals", "parallel"] = True
    if exe_name == "embed_enzo" or ("_parallel" in dir(sys) and sys._parallel):  # type: ignore
        ytcfg["yt", "inline"] = True
    yt.utilities.logger.uncolorize_logging()
    # Even though the uncolorize function already resets the format string,
    # we reset it again so that it includes the processor.
    f = logging.Formatter(
        "P%03i %s" % (communicator.rank, yt.utilities.logger.ufstring)
    )
    if len(yt.utilities.logger.ytLogger.handlers) > 0:
        yt.utilities.logger.ytLogger.handlers[0].setFormatter(f)

    if ytcfg.get("yt", "parallel_traceback"):
        sys.excepthook = traceback_writer_hook("_%03i" % communicator.rank)
    else:
        sys.excepthook = default_mpi_excepthook

    if ytcfg.get("yt", "log_level") < 20:
        yt.utilities.logger.ytLogger.warning(
            "Log Level is set low -- this could affect parallel performance!"
        )
    dtype_names.update(
        {
            "float32": MPI.FLOAT,
            "float64": MPI.DOUBLE,
            "int32": MPI.INT,
            "int64": MPI.LONG,
            "c": MPI.CHAR,
        }
    )
    op_names.update({"sum": MPI.SUM, "min": MPI.MIN, "max": MPI.MAX})
    # Turn off logging on all but the root rank, if specified.
    if suppress_logging:
        if communicator.rank > 0:
            mylog.addFilter(FilterAllMessages())
    return True


# Because the dtypes will == correctly but do not hash the same, we need this
# function for dictionary access.
def get_mpi_type(dtype):
    for dt, val in dtype_names.items():
        if dt == dtype:
            return val


class ObjectIterator:
    """
    This is a generalized class that accepts a list of objects and then
    attempts to intelligently iterate over them.
    """

    def __init__(self, pobj, just_list=False, attr="_grids"):
        self.pobj = pobj
        if hasattr(pobj, attr) and getattr(pobj, attr) is not None:
            gs = getattr(pobj, attr)
        else:
            gs = getattr(pobj._data_source, attr)
        if len(gs) == 0:
            raise YTNoDataInObjectError(pobj)
        if hasattr(gs[0], "proc_num"):
            # This one sort of knows about MPI, but not quite
            self._objs = [
                g
                for g in gs
                if g.proc_num == ytcfg.get("yt", "internals", "topcomm_parallel_rank")
            ]
            self._use_all = True
        else:
            self._objs = gs
            if hasattr(self._objs[0], "filename"):
                self._objs = sorted(self._objs, key=lambda g: g.filename)
            self._use_all = False
        self.ng = len(self._objs)
        self.just_list = just_list

    def __iter__(self):
        yield from self._objs


class ParallelObjectIterator(ObjectIterator):
    """
    This takes an object, *pobj*, that implements ParallelAnalysisInterface,
    and then does its thing, calling initialize and finalize on the object.
    """

    def __init__(self, pobj, just_list=False, attr="_grids", round_robin=False):
        ObjectIterator.__init__(self, pobj, just_list, attr=attr)
        # pobj has to be a ParallelAnalysisInterface, so it must have a .comm
        # object.
        self._offset = pobj.comm.rank
        self._skip = pobj.comm.size
        # Note that we're doing this in advance, and with a simple means
        # of choosing them; more advanced methods will be explored later.
        if self._use_all:
            self.my_obj_ids = np.arange(len(self._objs))
        else:
            if not round_robin:
                self.my_obj_ids = np.array_split(
                    np.arange(len(self._objs)), self._skip
                )[self._offset]
            else:
                self.my_obj_ids = np.arange(len(self._objs))[self._offset :: self._skip]

    def __iter__(self):
        for gid in self.my_obj_ids:
            yield self._objs[gid]
        if not self.just_list:
            self.pobj._finalize_parallel()


def parallel_simple_proxy(func):
    """
    This is a decorator that broadcasts the result of computation on a single
    processor to all other processors.  To do so, it uses the _processing and
    _distributed flags in the object to check for blocks.  Meant only to be
    used on objects that subclass
    :class:`~yt.utilities.parallel_tools.parallel_analysis_interface.ParallelAnalysisInterface`.
    """

    @wraps(func)
    def single_proc_results(self, *args, **kwargs):
        retval = None
        if hasattr(self, "dont_wrap"):
            if func.__name__ in self.dont_wrap:
                return func(self, *args, **kwargs)
        if not parallel_capable or self._processing or not self._distributed:
            return func(self, *args, **kwargs)
        comm = _get_comm((self,))
        if self._owner == comm.rank:
            self._processing = True
            retval = func(self, *args, **kwargs)
            self._processing = False
        # To be sure we utilize the root= kwarg, we manually access the .comm
        # attribute, which must be an instance of MPI.Intracomm, and call bcast
        # on that.
        retval = comm.comm.bcast(retval, root=self._owner)
        return retval

    return single_proc_results


class ParallelDummy(type):
    """
    This is a base class that, on instantiation, replaces all attributes that
    don't start with ``_`` with
    :func:`~yt.utilities.parallel_tools.parallel_analysis_interface.parallel_simple_proxy`-wrapped
    attributes.  Used as a metaclass.
    """

    def __init__(cls, name, bases, d):
        super().__init__(name, bases, d)
        skip = d.pop("dont_wrap", [])
        extra = d.pop("extra_wrap", [])
        for attrname in d:
            if attrname.startswith("_") or attrname in skip:
                if attrname not in extra:
                    continue
            attr = getattr(cls, attrname)
            if callable(attr):
                setattr(cls, attrname, parallel_simple_proxy(attr))


def parallel_passthrough(func):
    """
    If we are not run in parallel, this function passes the input back as
    output; otherwise, the function gets called.  Used as a decorator.
    """

    @wraps(func)
    def passage(self, *args, **kwargs):
        if not self._distributed:
            return args[0]
        return func(self, *args, **kwargs)

    return passage


def _get_comm(args):
    if len(args) > 0 and hasattr(args[0], "comm"):
        comm = args[0].comm
    else:
        comm = communication_system.communicators[-1]
    return comm


def parallel_blocking_call(func):
    """
    This decorator blocks on entry and exit of a function.
    """

    @wraps(func)
    def barrierize(*args, **kwargs):
        if not parallel_capable:
            return func(*args, **kwargs)
        mylog.debug("Entering barrier before %s", func.__name__)
        comm = _get_comm(args)
        comm.barrier()
        retval = func(*args, **kwargs)
        mylog.debug("Entering barrier after %s", func.__name__)
        comm.barrier()
        return retval

    return barrierize


def parallel_root_only(func):
    """
    This decorator blocks and calls the function on the root processor,
    but does not broadcast results to the other processors.
    """

    @wraps(func)
    def root_only(*args, **kwargs):
        if not parallel_capable:
            return func(*args, **kwargs)
        comm = _get_comm(args)
        rv = None
        if comm.rank == 0:
            try:
                rv = func(*args, **kwargs)
                all_clear = 1
            except Exception:
                traceback.print_last()
                all_clear = 0
        else:
            all_clear = None
        all_clear = comm.mpi_bcast(all_clear)
        if not all_clear:
            raise RuntimeError
        return rv

    return root_only


class Workgroup:
    def __init__(self, size, ranks, comm, name):
        self.size = size
        self.ranks = ranks
        self.comm = comm
        self.name = name


class ProcessorPool:
    comm = None
    size = None
    ranks = None
    available_ranks = None
    tasks = None

    def __init__(self):
        self.comm = communication_system.communicators[-1]
        self.size = self.comm.size
        self.ranks = list(range(self.size))
        self.available_ranks = list(range(self.size))
        self.workgroups = []

    def add_workgroup(self, size=None, ranks=None, name=None):
        if size is None:
            size = len(self.available_ranks)
        if len(self.available_ranks) < size:
            mylog.error(
                "Not enough resources available, asked for %d have %d",
                size,
                self.available_ranks,
            )
            raise RuntimeError
        if ranks is None:
            ranks = [self.available_ranks.pop(0) for i in range(size)]
        # Default name to the workgroup number.
        if name is None:
            name = str(len(self.workgroups))
        group = self.comm.comm.Get_group().Incl(ranks)
        new_comm = self.comm.comm.Create(group)
        if self.comm.rank in ranks:
            communication_system.communicators.append(Communicator(new_comm))
        self.workgroups.append(Workgroup(len(ranks), ranks, new_comm, name))

    def free_workgroup(self, workgroup):
        # If you want to actually delete the workgroup you will need to
        # pop it out of the self.workgroups list so you don't have references
        # that are left dangling, e.g. see free_all() below.
        for i in workgroup.ranks:
            if self.comm.rank == i:
                communication_system.communicators.pop()
            self.available_ranks.append(i)
        self.available_ranks.sort()

    def free_all(self):
        for wg in self.workgroups:
            self.free_workgroup(wg)
        while self.workgroups:
            self.workgroups.pop(0)

    @classmethod
    def from_sizes(cls, sizes):
        pool = cls()
        rank = pool.comm.rank
        for i, size in enumerate(always_iterable(sizes)):
            if is_sequence(size):
                size, name = size
            else:
                name = "workgroup_%02i" % i
            pool.add_workgroup(size, name=name)
        for wg in pool.workgroups:
            if rank in wg.ranks:
                workgroup = wg
        return pool, workgroup

    def __getitem__(self, key):
        for wg in self.workgroups:
            if wg.name == key:
                return wg
        raise KeyError(key)


class ResultsStorage:
    slots = ["result", "result_id"]
    result = None
    result_id = None


def parallel_objects(objects, njobs=0, storage=None, barrier=True, dynamic=False):
    r"""This function dispatches components of an iterable to different
    processors.

    The parallel_objects function accepts an iterable, *objects*, and based on
    the number of jobs requested and number of available processors, decides
    how to dispatch individual objects to processors or sets of processors.
    This can implicitly include multi-level parallelism, such that the
    processor groups assigned each object can be composed of several or even
    hundreds of processors.  *storage* is also available, for collation of
    results at the end of the iteration loop.

    Calls to this function can be nested.

    This should not be used to iterate over datasets --
    :class:`~yt.data_objects.time_series.DatasetSeries` provides a much nicer
    interface for that.

    Parameters
    ----------
    objects : Iterable
        The list of objects to dispatch to different processors.
    njobs : int
        How many jobs to spawn.  By default, one job will be dispatched for
        each available processor.
    storage : dict
        This is a dictionary, which will be filled with results during the
        course of the iteration.  The keys will be the dataset
        indices and the values will be whatever is assigned to the *result*
        attribute on the storage during iteration.
    barrier : bool
        Should a barier be placed at the end of iteration?
    dynamic : bool
        This governs whether or not dynamic load balancing will be enabled.
        This requires one dedicated processor; if this is enabled with a set of
        128 processors available, only 127 will be available to iterate over
        objects as one will be load balancing the rest.


    Examples
    --------
    Here is a simple example of iterating over a set of centers and making
    slice plots centered at each.

    >>> for c in parallel_objects(centers):
    ...     SlicePlot(ds, "x", "Density", center=c).save()
    ...

    Here's an example of calculating the angular momentum vector of a set of
    spheres, but with a set of four jobs of multiple processors each.  Note
    that we also store the results.

    >>> storage = {}
    >>> for sto, c in parallel_objects(centers, njobs=4, storage=storage):
    ...     sp = ds.sphere(c, (100, "kpc"))
    ...     sto.result = sp.quantities["AngularMomentumVector"]()
    ...
    >>> for sphere_id, L in sorted(storage.items()):
    ...     print(centers[sphere_id], L)
    ...

    """
    if dynamic:
        from .task_queue import dynamic_parallel_objects

        yield from dynamic_parallel_objects(objects, njobs=njobs, storage=storage)
        return

    if not parallel_capable:
        njobs = 1
    my_communicator = communication_system.communicators[-1]
    my_size = my_communicator.size
    if njobs <= 0:
        njobs = my_size
    if njobs > my_size:
        mylog.error(
            "You have asked for %s jobs, but you only have %s processors.",
            njobs,
            my_size,
        )
        raise RuntimeError
    my_rank = my_communicator.rank
    all_new_comms = np.array_split(np.arange(my_size), njobs)
    for i, comm_set in enumerate(all_new_comms):
        if my_rank in comm_set:
            my_new_id = i
            break
    if parallel_capable:
        communication_system.push_with_ids(all_new_comms[my_new_id].tolist())
    to_share = {}
    # If our objects object is slice-aware, like time series data objects are,
    # this will prevent intermediate objects from being created.
    oiter = itertools.islice(enumerate(objects), my_new_id, None, njobs)
    for result_id, obj in oiter:
        if storage is not None:
            rstore = ResultsStorage()
            rstore.result_id = result_id
            yield rstore, obj
            to_share[rstore.result_id] = rstore.result
        else:
            yield obj
    if parallel_capable:
        communication_system.pop()
    if storage is not None:
        # Now we have to broadcast it
        new_storage = my_communicator.par_combine_object(
            to_share, datatype="dict", op="join"
        )
        storage.update(new_storage)
    if barrier:
        my_communicator.barrier()


def parallel_ring(objects, generator_func, mutable=False):
    r"""This function loops in a ring around a set of objects, yielding the
    results of generator_func and passing from one processor to another to
    avoid IO or expensive computation.

    This function is designed to operate in sequence on a set of objects, where
    the creation of those objects might be expensive.  For instance, this could
    be a set of particles that are costly to read from disk.  Processor N will
    run generator_func on an object, and the results of that will both be
    yielded and passed to processor N-1.  If the length of the objects is not
    equal to the number of processors, then the final processor in the top
    communicator will re-generate the data as needed.

    In all likelihood, this function will only be useful internally to yt.

    Parameters
    ----------
    objects : Iterable
        The list of objects to operate on.
    generator_func : callable
        This function will be called on each object, and the results yielded.
        It must return a single NumPy array; for multiple values, it needs to
        have a custom dtype.
    mutable : bool
        Should the arrays be considered mutable?  Currently, this will only
        work if the number of processors equals the number of objects.

    Examples
    --------
    Here is a simple example of a ring loop around a set of integers, with a
    custom dtype.

    >>> dt = np.dtype([("x", "float64"), ("y", "float64"), ("z", "float64")])
    >>> def gfunc(o):
    ...     np.random.seed(o)
    ...     rv = np.empty(1000, dtype=dt)
    ...     rv["x"] = np.random.random(1000)
    ...     rv["y"] = np.random.random(1000)
    ...     rv["z"] = np.random.random(1000)
    ...     return rv
    ...
    >>> obj = range(8)
    >>> for obj, arr in parallel_ring(obj, gfunc):
    ...     print(arr["x"].sum(), arr["y"].sum(), arr["z"].sum())
    ...

    """
    if mutable:
        raise NotImplementedError
    my_comm = communication_system.communicators[-1]
    my_size = my_comm.size
    my_rank = my_comm.rank  # This will also be the first object we access
    if not parallel_capable and not mutable:
        for obj in objects:
            yield obj, generator_func(obj)
        return
    generate_endpoints = len(objects) != my_size
    # gback False: send the object backwards
    # gforw False: receive an object from forwards
    if len(objects) == my_size:
        generate_endpoints = False
        gback = False
        gforw = False
    else:
        # In this case, the first processor (my_rank == 0) will generate.
        generate_endpoints = True
        gback = my_rank == 0
        gforw = my_rank == my_size - 1
    if generate_endpoints and mutable:
        raise NotImplementedError
    # Now we need to do pairwise sends
    source = (my_rank + 1) % my_size
    dest = (my_rank - 1) % my_size
    oiter = itertools.islice(itertools.cycle(objects), my_rank, my_rank + len(objects))
    idata = None
    isize = np.zeros((1,), dtype="int64")
    osize = np.zeros((1,), dtype="int64")
    for obj in oiter:
        if idata is None or gforw:
            idata = generator_func(obj)
            idtype = odtype = idata.dtype
            if get_mpi_type(idtype) is None:
                idtype = "c"
        yield obj, idata
        # We first send to the previous processor
        tags = []
        if not gforw:
            tags.append(my_comm.mpi_nonblocking_recv(isize, source))
        if not gback:
            osize[0] = idata.size
            tags.append(my_comm.mpi_nonblocking_send(osize, dest))
        my_comm.mpi_Request_Waitall(tags)
        odata = idata
        tags = []
        if not gforw:
            idata = np.empty(isize[0], dtype=odtype)
            tags.append(
                my_comm.mpi_nonblocking_recv(idata.view(idtype), source, dtype=idtype)
            )
        if not gback:
            tags.append(
                my_comm.mpi_nonblocking_send(odata.view(idtype), dest, dtype=idtype)
            )
        my_comm.mpi_Request_Waitall(tags)
        del odata


class CommunicationSystem:
    communicators: List["Communicator"] = []

    def __init__(self):
        self.communicators.append(Communicator(None))

    def push(self, new_comm):
        if not isinstance(new_comm, Communicator):
            new_comm = Communicator(new_comm)
        self.communicators.append(new_comm)
        self._update_parallel_state(new_comm)

    def push_with_ids(self, ids):
        group = self.communicators[-1].comm.Get_group().Incl(ids)
        new_comm = self.communicators[-1].comm.Create(group)
        self.push(new_comm)
        return new_comm

    def _update_parallel_state(self, new_comm):
        ytcfg["yt", "internals", "topcomm_parallel_size"] = new_comm.size
        ytcfg["yt", "internals", "topcomm_parallel_rank"] = new_comm.rank
        if new_comm.rank > 0 and ytcfg.get("yt", "serialize"):
            ytcfg["yt", "only_deserialize"] = True

    def pop(self):
        self.communicators.pop()
        self._update_parallel_state(self.communicators[-1])


def _reconstruct_communicator():
    return communication_system.communicators[-1]


class Communicator:
    comm = None
    _grids = None
    _distributed = None
    __tocast = "c"

    def __init__(self, comm=None):
        self.comm = comm
        self._distributed = comm is not None and self.comm.size > 1

    def __del__(self):
        if self.comm is not None:
            self.comm.Free()

    """
    This is an interface specification providing several useful utility
    functions for analyzing something in parallel.
    """

    def __reduce__(self):
        # We don't try to reconstruct any of the properties of the communicator
        # or the processors.  In general, we don't want to.
        return (_reconstruct_communicator, ())

    def barrier(self):
        if not self._distributed:
            return
        mylog.debug("Opening MPI Barrier on %s", self.comm.rank)
        self.comm.Barrier()

    def mpi_exit_test(self, data=False):
        # data==True -> exit. data==False -> no exit
        mine, statuses = self.mpi_info_dict(data)
        if True in statuses.values():
            raise RuntimeError("Fatal error. Exiting.")
        return None

    @parallel_passthrough
    def par_combine_object(self, data, op, datatype=None):
        # op can be chosen from:
        #   cat
        #   join
        # data is selected to be of types:
        #   np.ndarray
        #   dict
        #   data field dict
        if datatype is not None:
            pass
        elif isinstance(data, dict):
            datatype = "dict"
        elif isinstance(data, np.ndarray):
            datatype = "array"
        elif isinstance(data, list):
            datatype = "list"
        # Now we have our datatype, and we conduct our operation
        if datatype == "dict" and op == "join":
            if self.comm.rank == 0:
                for i in range(1, self.comm.size):
                    data.update(self.comm.recv(source=i, tag=0))
            else:
                self.comm.send(data, dest=0, tag=0)

            # Send the keys first, then each item one by one
            # This is to prevent MPI from crashing when sending more
            # than 2GiB of data over the network.
            keys = self.comm.bcast(list(data.keys()), root=0)
            for key in keys:
                tmp = data.get(key, None)
                data[key] = self.comm.bcast(tmp, root=0)
            return data
        elif datatype == "dict" and op == "cat":
            field_keys = sorted(data.keys())
            size = data[field_keys[0]].shape[-1]
            sizes = np.zeros(self.comm.size, dtype="int64")
            outsize = np.array(size, dtype="int64")
            self.comm.Allgather([outsize, 1, MPI.LONG], [sizes, 1, MPI.LONG])
            # This nested concatenate is to get the shapes to work out correctly;
            # if we just add [0] to sizes, it will broadcast a summation, not a
            # concatenation.
            offsets = np.add.accumulate(np.concatenate([[0], sizes]))[:-1]
            arr_size = self.comm.allreduce(size, op=MPI.SUM)
            for key in field_keys:
                dd = data[key]
                rv = self.alltoallv_array(dd, arr_size, offsets, sizes)
                data[key] = rv
            return data
        elif datatype == "array" and op == "cat":
            if data is None:
                ncols = -1
                size = 0
                dtype = "float64"
                mylog.warning(
                    "Array passed to par_combine_object was None. "
                    "Setting dtype to float64. This may break things!"
                )
            else:
                dtype = data.dtype
                if len(data) == 0:
                    ncols = -1
                    size = 0
                elif len(data.shape) == 1:
                    ncols = 1
                    size = data.shape[0]
                else:
                    ncols, size = data.shape
            ncols = self.comm.allreduce(ncols, op=MPI.MAX)
            if ncols == 0:
                data = np.zeros(0, dtype=dtype)  # This only works for
            elif data is None:
                data = np.zeros((ncols, 0), dtype=dtype)
            size = data.shape[-1]
            sizes = np.zeros(self.comm.size, dtype="int64")
            outsize = np.array(size, dtype="int64")
            self.comm.Allgather([outsize, 1, MPI.LONG], [sizes, 1, MPI.LONG])
            # This nested concatenate is to get the shapes to work out correctly;
            # if we just add [0] to sizes, it will broadcast a summation, not a
            # concatenation.
            offsets = np.add.accumulate(np.concatenate([[0], sizes]))[:-1]
            arr_size = self.comm.allreduce(size, op=MPI.SUM)
            data = self.alltoallv_array(data, arr_size, offsets, sizes)
            return data
        elif datatype == "list" and op == "cat":
            recv_data = self.comm.allgather(data)
            # Now flatten into a single list, since this
            # returns us a list of lists.
            data = []
            while recv_data:
                data.extend(recv_data.pop(0))
            return data
        raise NotImplementedError

    @parallel_passthrough
    def mpi_bcast(self, data, root=0):
        # The second check below makes sure that we know how to communicate
        # this type of array. Otherwise, we'll pickle it.
        if isinstance(data, np.ndarray) and get_mpi_type(data.dtype) is not None:
            if self.comm.rank == root:
                if isinstance(data, YTArray):
                    info = (
                        data.shape,
                        data.dtype,
                        str(data.units),
                        data.units.registry.lut,
                    )
                    if isinstance(data, ImageArray):
                        info += ("ImageArray",)
                    else:
                        info += ("YTArray",)
                else:
                    info = (data.shape, data.dtype)
            else:
                info = ()
            info = self.comm.bcast(info, root=root)
            if self.comm.rank != root:
                if len(info) == 5:
                    registry = UnitRegistry(lut=info[3], add_default_symbols=False)
                    if info[-1] == "ImageArray":
                        data = ImageArray(
                            np.empty(info[0], dtype=info[1]),
                            units=info[2],
                            registry=registry,
                        )
                    else:
                        data = YTArray(
                            np.empty(info[0], dtype=info[1]), info[2], registry=registry
                        )
                else:
                    data = np.empty(info[0], dtype=info[1])
            mpi_type = get_mpi_type(info[1])
            self.comm.Bcast([data, mpi_type], root=root)
            return data
        else:
            # Use pickled methods.
            data = self.comm.bcast(data, root=root)
            return data

    def preload(self, grids, fields, io_handler):
        # This is non-functional.
        return

    @parallel_passthrough
    def mpi_allreduce(self, data, dtype=None, op="sum"):
        op = op_names[op]
        if isinstance(data, np.ndarray) and data.dtype != bool:
            if dtype is None:
                dtype = data.dtype
            if dtype != data.dtype:
                data = data.astype(dtype)
            temp = data.copy()
            self.comm.Allreduce(
                [temp, get_mpi_type(dtype)], [data, get_mpi_type(dtype)], op
            )
            return data
        else:
            # We use old-school pickling here on the assumption the arrays are
            # relatively small ( < 1e7 elements )
            return self.comm.allreduce(data, op)

    ###
    # Non-blocking stuff.
    ###

    def mpi_nonblocking_recv(self, data, source, tag=0, dtype=None):
        if not self._distributed:
            return -1
        if dtype is None:
            dtype = data.dtype
        mpi_type = get_mpi_type(dtype)
        return self.comm.Irecv([data, mpi_type], source, tag)

    def mpi_nonblocking_send(self, data, dest, tag=0, dtype=None):
        if not self._distributed:
            return -1
        if dtype is None:
            dtype = data.dtype
        mpi_type = get_mpi_type(dtype)
        return self.comm.Isend([data, mpi_type], dest, tag)

    def mpi_Request_Waitall(self, hooks):
        if not self._distributed:
            return
        MPI.Request.Waitall(hooks)

    def mpi_Request_Waititer(self, hooks):
        for _hook in hooks:
            req = MPI.Request.Waitany(hooks)
            yield req

    def mpi_Request_Testall(self, hooks):
        """
        This returns False if any of the request hooks are un-finished,
        and True if they are all finished.
        """
        if not self._distributed:
            return True
        return MPI.Request.Testall(hooks)

    ###
    # End non-blocking stuff.
    ###

    ###
    # Parallel rank and size properties.
    ###

    @property
    def size(self):
        if not self._distributed:
            return 1
        return self.comm.size

    @property
    def rank(self):
        if not self._distributed:
            return 0
        return self.comm.rank

    def mpi_info_dict(self, info):
        if not self._distributed:
            return 0, {0: info}
        data = None
        if self.comm.rank == 0:
            data = {0: info}
            for i in range(1, self.comm.size):
                data[i] = self.comm.recv(source=i, tag=0)
        else:
            self.comm.send(info, dest=0, tag=0)
        mylog.debug("Opening MPI Broadcast on %s", self.comm.rank)
        data = self.comm.bcast(data, root=0)
        return self.comm.rank, data

    def claim_object(self, obj):
        if not self._distributed:
            return
        obj._owner = self.comm.rank
        obj._distributed = True

    def do_not_claim_object(self, obj):
        if not self._distributed:
            return
        obj._owner = -1
        obj._distributed = True

    def write_on_root(self, fn):
        if not self._distributed:
            return open(fn, "w")
        if self.comm.rank == 0:
            return open(fn, "w")
        else:
            return StringIO()

    def get_filename(self, prefix, rank=None):
        if not self._distributed:
            return prefix
        if rank is None:
            return "%s_%04i" % (prefix, self.comm.rank)
        else:
            return "%s_%04i" % (prefix, rank)

    def is_mine(self, obj):
        if not obj._distributed:
            return True
        return obj._owner == self.comm.rank

    def send_quadtree(self, target, buf, tgd, args):
        sizebuf = np.zeros(1, "int64")
        sizebuf[0] = buf[0].size
        self.comm.Send([sizebuf, MPI.LONG], dest=target)
        self.comm.Send([buf[0], MPI.INT], dest=target)
        self.comm.Send([buf[1], MPI.DOUBLE], dest=target)
        self.comm.Send([buf[2], MPI.DOUBLE], dest=target)

    def recv_quadtree(self, target, tgd, args):
        sizebuf = np.zeros(1, "int64")
        self.comm.Recv(sizebuf, source=target)
        buf = [
            np.empty((sizebuf[0],), "int32"),
            np.empty((sizebuf[0], args[2]), "float64"),
            np.empty((sizebuf[0],), "float64"),
        ]
        self.comm.Recv([buf[0], MPI.INT], source=target)
        self.comm.Recv([buf[1], MPI.DOUBLE], source=target)
        self.comm.Recv([buf[2], MPI.DOUBLE], source=target)
        return buf

    @parallel_passthrough
    def merge_quadtree_buffers(self, qt, merge_style):
        # This is a modified version of pairwise reduction from Lisandro Dalcin,
        # in the reductions demo of mpi4py
        size = self.comm.size
        rank = self.comm.rank

        mask = 1

        buf = qt.tobuffer()
        print("PROC", rank, buf[0].shape, buf[1].shape, buf[2].shape)
        sys.exit()

        args = qt.get_args()  # Will always be the same
        tgd = np.array([args[0], args[1]], dtype="int64")
        sizebuf = np.zeros(1, "int64")

        while mask < size:
            if (mask & rank) != 0:
                target = (rank & ~mask) % size
                # print("SENDING FROM %02i to %02i" % (rank, target))
                buf = qt.tobuffer()
                self.send_quadtree(target, buf, tgd, args)
                # qt = self.recv_quadtree(target, tgd, args)
            else:
                target = rank | mask
                if target < size:
                    # print("RECEIVING FROM %02i on %02i" % (target, rank))
                    buf = self.recv_quadtree(target, tgd, args)
                    qto = QuadTree(tgd, args[2], qt.bounds)
                    qto.frombuffer(buf[0], buf[1], buf[2], merge_style)
                    merge_quadtrees(qt, qto, style=merge_style)
                    del qto
                    # self.send_quadtree(target, qt, tgd, args)
            mask <<= 1

        if rank == 0:
            buf = qt.tobuffer()
            sizebuf[0] = buf[0].size
        self.comm.Bcast([sizebuf, MPI.LONG], root=0)
        if rank != 0:
            buf = [
                np.empty((sizebuf[0],), "int32"),
                np.empty((sizebuf[0], args[2]), "float64"),
                np.empty((sizebuf[0],), "float64"),
            ]
        self.comm.Bcast([buf[0], MPI.INT], root=0)
        self.comm.Bcast([buf[1], MPI.DOUBLE], root=0)
        self.comm.Bcast([buf[2], MPI.DOUBLE], root=0)
        self.refined = buf[0]
        if rank != 0:
            qt = QuadTree(tgd, args[2], qt.bounds)
            qt.frombuffer(buf[0], buf[1], buf[2], merge_style)
        return qt

    def send_array(self, arr, dest, tag=0):
        if not isinstance(arr, np.ndarray):
            self.comm.send((None, None), dest=dest, tag=tag)
            self.comm.send(arr, dest=dest, tag=tag)
            return
        tmp = arr.view(self.__tocast)  # Cast to CHAR
        # communicate type and shape and optionally units
        if isinstance(arr, YTArray):
            unit_metadata = (str(arr.units), arr.units.registry.lut)
            if isinstance(arr, ImageArray):
                unit_metadata += ("ImageArray",)
            else:
                unit_metadata += ("YTArray",)
        else:
            unit_metadata = ()
        self.comm.send((arr.dtype.str, arr.shape) + unit_metadata, dest=dest, tag=tag)
        self.comm.Send([arr, MPI.CHAR], dest=dest, tag=tag)
        del tmp

    def recv_array(self, source, tag=0):
        metadata = self.comm.recv(source=source, tag=tag)
        dt, ne = metadata[:2]
        if ne is None and dt is None:
            return self.comm.recv(source=source, tag=tag)
        arr = np.empty(ne, dtype=dt)
        if len(metadata) == 5:
            registry = UnitRegistry(lut=metadata[3], add_default_symbols=False)
            if metadata[-1] == "ImageArray":
                arr = ImageArray(arr, units=metadata[2], registry=registry)
            else:
                arr = YTArray(arr, metadata[2], registry=registry)
        tmp = arr.view(self.__tocast)
        self.comm.Recv([tmp, MPI.CHAR], source=source, tag=tag)
        return arr

    def alltoallv_array(self, send, total_size, offsets, sizes):
        if len(send.shape) > 1:
            recv = []
            for i in range(send.shape[0]):
                recv.append(
                    self.alltoallv_array(send[i, :].copy(), total_size, offsets, sizes)
                )
            recv = np.array(recv)
            return recv
        offset = offsets[self.comm.rank]
        tmp_send = send.view(self.__tocast)
        recv = np.empty(total_size, dtype=send.dtype)
        if isinstance(send, YTArray):
            # We assume send.units is consistent with the units
            # on the receiving end.
            if isinstance(send, ImageArray):
                recv = ImageArray(recv, units=send.units)
            else:
                recv = YTArray(recv, send.units)
        recv[offset : offset + send.size] = send[:]
        dtr = send.dtype.itemsize / tmp_send.dtype.itemsize  # > 1
        roff = [off * dtr for off in offsets]
        rsize = [siz * dtr for siz in sizes]
        tmp_recv = recv.view(self.__tocast)
        self.comm.Allgatherv(
            (tmp_send, tmp_send.size, MPI.CHAR), (tmp_recv, (rsize, roff), MPI.CHAR)
        )
        return recv

    def probe_loop(self, tag, callback):
        while True:
            st = MPI.Status()
            self.comm.Probe(MPI.ANY_SOURCE, tag=tag, status=st)
            try:
                callback(st)
            except StopIteration:
                mylog.debug("Probe loop ending.")
                break


communication_system = CommunicationSystem()


class ParallelAnalysisInterface:
    comm = None
    _grids = None
    _distributed = None

    def __init__(self, comm=None):
        if comm is None:
            self.comm = communication_system.communicators[-1]
        else:
            self.comm = comm
        self._grids = self.comm._grids
        self._distributed = self.comm._distributed

    def _get_objs(self, attr, *args, **kwargs):
        if self._distributed:
            rr = kwargs.pop("round_robin", False)
            self._initialize_parallel(*args, **kwargs)
            return ParallelObjectIterator(self, attr=attr, round_robin=rr)
        return ObjectIterator(self, attr=attr)

    def _get_grids(self, *args, **kwargs):
        if self._distributed:
            self._initialize_parallel(*args, **kwargs)
            return ParallelObjectIterator(self, attr="_grids")
        return ObjectIterator(self, attr="_grids")

    def _get_grid_objs(self):
        if self._distributed:
            return ParallelObjectIterator(self, True, attr="_grids")
        return ObjectIterator(self, True, attr="_grids")

    def get_dependencies(self, fields):
        deps = []
        fi = self.ds.field_info
        for field in fields:
            if any(getattr(v, "ghost_zones", 0) > 0 for v in fi[field].validators):
                continue
            deps += list(
                always_iterable(fi[field].get_dependencies(ds=self.ds).requested)
            )
        return list(set(deps))

    def _initialize_parallel(self):
        pass

    def _finalize_parallel(self):
        pass

    def partition_index_2d(self, axis):
        if not self._distributed:
            return False, self.index.grid_collection(self.center, self.index.grids)

        xax = self.ds.coordinates.x_axis[axis]
        yax = self.ds.coordinates.y_axis[axis]
        cc = MPI.Compute_dims(self.comm.size, 2)
        mi = self.comm.rank
        cx, cy = np.unravel_index(mi, cc)
        x = np.mgrid[0 : 1 : (cc[0] + 1) * 1j][cx : cx + 2]
        y = np.mgrid[0 : 1 : (cc[1] + 1) * 1j][cy : cy + 2]

        DLE, DRE = self.ds.domain_left_edge.copy(), self.ds.domain_right_edge.copy()
        LE = np.ones(3, dtype="float64") * DLE
        RE = np.ones(3, dtype="float64") * DRE
        LE[xax] = x[0] * (DRE[xax] - DLE[xax]) + DLE[xax]
        RE[xax] = x[1] * (DRE[xax] - DLE[xax]) + DLE[xax]
        LE[yax] = y[0] * (DRE[yax] - DLE[yax]) + DLE[yax]
        RE[yax] = y[1] * (DRE[yax] - DLE[yax]) + DLE[yax]
        mylog.debug("Dimensions: %s %s", LE, RE)

        reg = self.ds.region(self.center, LE, RE)
        return True, reg

    def partition_index_3d(self, ds, padding=0.0, rank_ratio=1):
        LE, RE = np.array(ds.left_edge), np.array(ds.right_edge)
        # We need to establish if we're looking at a subvolume, in which case
        # we *do* want to pad things.
        if (LE == self.ds.domain_left_edge).all() and (
            RE == self.ds.domain_right_edge
        ).all():
            subvol = False
        else:
            subvol = True
        if not self._distributed and not subvol:
            return False, LE, RE, ds
        if not self._distributed and subvol:
            return True, LE, RE, self.ds.region(self.center, LE - padding, RE + padding)
        elif ytcfg.get("yt", "inline"):
            # At this point, we want to identify the root grid tile to which
            # this processor is assigned.
            # The only way I really know how to do this is to get the level-0
            # grid that belongs to this processor.
            grids = self.ds.index.select_grids(0)
            root_grids = [g for g in grids if g.proc_num == self.comm.rank]
            if len(root_grids) != 1:
                raise RuntimeError
            # raise KeyError
            LE = root_grids[0].LeftEdge
            RE = root_grids[0].RightEdge
            return True, LE, RE, self.ds.region(self.center, LE, RE)

        cc = MPI.Compute_dims(self.comm.size / rank_ratio, 3)
        mi = self.comm.rank % (self.comm.size // rank_ratio)
        cx, cy, cz = np.unravel_index(mi, cc)
        x = np.mgrid[LE[0] : RE[0] : (cc[0] + 1) * 1j][cx : cx + 2]
        y = np.mgrid[LE[1] : RE[1] : (cc[1] + 1) * 1j][cy : cy + 2]
        z = np.mgrid[LE[2] : RE[2] : (cc[2] + 1) * 1j][cz : cz + 2]

        LE = np.array([x[0], y[0], z[0]], dtype="float64")
        RE = np.array([x[1], y[1], z[1]], dtype="float64")

        if padding > 0:
            return True, LE, RE, self.ds.region(self.center, LE - padding, RE + padding)

        return False, LE, RE, self.ds.region(self.center, LE, RE)

    def partition_region_3d(self, left_edge, right_edge, padding=0.0, rank_ratio=1):
        """
        Given a region, it subdivides it into smaller regions for parallel
        analysis.
        """
        LE, RE = left_edge[:], right_edge[:]
        if not self._distributed:
            raise NotImplementedError

        cc = MPI.Compute_dims(self.comm.size / rank_ratio, 3)
        mi = self.comm.rank % (self.comm.size // rank_ratio)
        cx, cy, cz = np.unravel_index(mi, cc)
        x = np.mgrid[LE[0] : RE[0] : (cc[0] + 1) * 1j][cx : cx + 2]
        y = np.mgrid[LE[1] : RE[1] : (cc[1] + 1) * 1j][cy : cy + 2]
        z = np.mgrid[LE[2] : RE[2] : (cc[2] + 1) * 1j][cz : cz + 2]

        LE = np.array([x[0], y[0], z[0]], dtype="float64")
        RE = np.array([x[1], y[1], z[1]], dtype="float64")

        if padding > 0:
            return True, LE, RE, self.ds.region(self.center, LE - padding, RE + padding)

        return False, LE, RE, self.ds.region(self.center, LE, RE)

    def partition_index_3d_bisection_list(self):
        """
        Returns an array that is used to drive _partition_index_3d_bisection,
        below.
        """

        def factor(n):
            if n == 1:
                return [1]
            i = 2
            limit = n**0.5
            while i <= limit:
                if n % i == 0:
                    ret = factor(n / i)
                    ret.append(i)
                    return ret
                i += 1
            return [n]

        cc = MPI.Compute_dims(self.comm.size, 3)
        si = self.comm.size

        factors = factor(si)
        xyzfactors = [factor(cc[0]), factor(cc[1]), factor(cc[2])]

        # Each entry of cuts is a two element list, that is:
        # [cut dim, number of cuts]
        cuts = []
        # The higher cuts are in the beginning.
        # We're going to do our best to make the cuts cyclic, i.e. x, then y,
        # then z, etc...
        lastdim = 0
        for f in factors:
            nextdim = (lastdim + 1) % 3
            while True:
                if f in xyzfactors[nextdim]:
                    cuts.append([nextdim, f])
                    topop = xyzfactors[nextdim].index(f)
                    xyzfactors[nextdim].pop(topop)
                    lastdim = nextdim
                    break
                nextdim = (nextdim + 1) % 3
        return cuts


class GroupOwnership(ParallelAnalysisInterface):
    def __init__(self, items):
        ParallelAnalysisInterface.__init__(self)
        self.num_items = len(items)
        self.items = items
        assert self.num_items >= self.comm.size
        self.owned = list(range(self.comm.size))
        self.pointer = 0
        if parallel_capable:
            communication_system.push_with_ids([self.comm.rank])

    def __del__(self):
        if parallel_capable:
            communication_system.pop()

    def inc(self, n=-1):
        old_item = self.item
        if n == -1:
            n = self.comm.size
        for _ in range(n):
            if self.pointer >= self.num_items - self.comm.size:
                break
            self.owned[self.pointer % self.comm.size] += self.comm.size
            self.pointer += 1
        if self.item is not old_item:
            self.switch()

    def dec(self, n=-1):
        old_item = self.item
        if n == -1:
            n = self.comm.size
        for _ in range(n):
            if self.pointer == 0:
                break
            self.owned[(self.pointer - 1) % self.comm.size] -= self.comm.size
            self.pointer -= 1
        if self.item is not old_item:
            self.switch()

    _last = None

    @property
    def item(self):
        own = self.owned[self.comm.rank]
        if self._last != own:
            self._item = self.items[own]
            self._last = own
        return self._item

    def switch(self):
        pass



# distutils: language = c++
# distutils: extra_compile_args = CPP14_FLAG OMP_ARGS
"""This is a wrapper around the C++ class to efficiently cast rays into an octree.
It relies on the seminal paper by  J. Revelles,, C.Ureña and M.Lastra.
"""


cimport numpy as np
import numpy as np

cimport cython


cdef class _OctreeRayTracing:
    def __init__(self, np.ndarray LE, np.ndarray RE, int depth):
        cdef double* LE_ptr = <double *>LE.data
        cdef double* RE_ptr = <double *>RE.data
        self.oct = new Octree[int](depth, LE_ptr, RE_ptr)
        self.depth = depth

    @cython.boundscheck(False)
    @cython.wraparound(False)
    def add_nodes(self, int[:, :] ipos_view, int[:] lvl_view, int[:] key):
        cdef int i
        cdef int ii[3]

        for i in range(len(key)):
            ii[0] = ipos_view[i, 0]
            ii[1] = ipos_view[i, 1]
            ii[2] = ipos_view[i, 2]
            self.oct.insert_node_no_ret(ii, lvl_view[i], <int> key[i])

    def __dealloc__(self):
        del self.oct



from nose.tools import assert_raises
from numpy.testing import assert_equal

from yt.funcs import just_one, levenshtein_distance, validate_axis, validate_center
from yt.testing import fake_amr_ds
from yt.units import YTArray, YTQuantity


def test_validate_axis():
    validate_axis(None, 0)
    validate_axis(None, "X")

    ds = fake_amr_ds(geometry="cylindrical")
    ds.slice("Theta", 0.25)

    with assert_raises(TypeError) as ex:
        # default geometry is cartesian
        ds = fake_amr_ds()
        ds.slice("r", 0.25)
    desired = "Expected axis to be any of [0, 1, 2, 'x', 'y', 'z', 'X', 'Y', 'Z'], received 'r'"

    actual = str(ex.exception)
    assert actual == desired


def test_validate_center():
    validate_center("max")
    validate_center("MIN_")

    with assert_raises(TypeError) as ex:
        validate_center("avg")
    desired = (
        "Expected 'center' to be in ['c', 'center', 'm', 'max', 'min'] "
        "or the prefix to be 'max_'/'min_', received 'avg'."
    )
    assert_equal(str(ex.exception), desired)

    validate_center(YTQuantity(0.25, "cm"))
    validate_center([0.25, 0.25, 0.25])

    class CustomCenter:
        def __init__(self, center):
            self.center = center

    with assert_raises(TypeError) as ex:
        validate_center(CustomCenter(10))
    desired = (
        "Expected 'center' to be a numeric object of type "
        "list/tuple/np.ndarray/YTArray/YTQuantity, received "
        "'yt.tests.test_funcs.test_validate_center.<locals>."
        "CustomCenter'."
    )
    assert_equal(str(ex.exception)[:50], desired[:50])


def test_just_one():
    # Check that behaviour of this function is consistent before and after refactor
    # PR 2893
    for unit in ["mm", "cm", "km", "pc", "g", "kg", "M_sun"]:
        obj = YTArray([0.0, 1.0], unit)
        expected = YTQuantity(obj.flat[0], obj.units, registry=obj.units.registry)
        jo = just_one(obj)
        assert jo == expected


def test_levenshtein():
    assert_equal(levenshtein_distance("abcdef", "abcdef"), 0)

    # Deletions / additions
    assert_equal(levenshtein_distance("abcdef", "abcde"), 1)
    assert_equal(levenshtein_distance("abcdef", "abcd"), 2)
    assert_equal(levenshtein_distance("abcdef", "abc"), 3)

    assert_equal(levenshtein_distance("abcdf", "abcdef"), 1)
    assert_equal(levenshtein_distance("cdef", "abcdef"), 2)
    assert_equal(levenshtein_distance("bde", "abcdef"), 3)

    # Substitutions
    assert_equal(levenshtein_distance("abcd", "abc_"), 1)
    assert_equal(levenshtein_distance("abcd", "ab__"), 2)
    assert_equal(levenshtein_distance("abcd", "a___"), 3)
    assert_equal(levenshtein_distance("abcd", "____"), 4)

    # Deletion + Substitutions
    assert_equal(levenshtein_distance("abcd", "abc_z"), 2)
    assert_equal(levenshtein_distance("abcd", "ab__zz"), 4)
    assert_equal(levenshtein_distance("abcd", "a___zzz"), 6)
    assert_equal(levenshtein_distance("abcd", "____zzzz"), 8)

    # Max distance
    assert_equal(levenshtein_distance("abcd", "", max_dist=0), 1)
    assert_equal(levenshtein_distance("abcd", "", max_dist=3), 4)
    assert_equal(levenshtein_distance("abcd", "", max_dist=10), 4)
    assert_equal(levenshtein_distance("abcd", "", max_dist=1), 2)
    assert_equal(levenshtein_distance("abcd", "a", max_dist=2), 3)
    assert_equal(levenshtein_distance("abcd", "ad", max_dist=2), 2)
    assert_equal(levenshtein_distance("abcd", "abd", max_dist=2), 1)
    assert_equal(levenshtein_distance("abcd", "abcd", max_dist=2), 0)



import pickle

from numpy.testing import assert_equal

from yt.testing import requires_file, requires_module
from yt.utilities.answer_testing.framework import data_dir_load

tipsy_ds = "TipsyGalaxy/galaxy.00300"
enzo_ds = "enzo_tiny_cosmology/DD0000/DD0000"


@requires_module("h5py")
@requires_file(enzo_ds)
def test_grid_pickles():
    ds = data_dir_load(enzo_ds)
    ad = ds.all_data()
    # just test ad since there is a nested ds that will get (un)pickled
    ad_pickle = pickle.loads(pickle.dumps(ad))
    assert_equal(ad.ds.basename, ad_pickle.ds.basename)


@requires_file(tipsy_ds)
def test_particle_pickles():
    ds = data_dir_load(tipsy_ds)
    ad = ds.all_data()
    ds.index._identify_base_chunk(ad)
    ch0 = list(ds.index._chunk_io(ad, cache=False))[0]
    # just test one chunk since there is a nested ds that will get (un)pickled
    ch_pickle = pickle.loads(pickle.dumps(ch0))
    assert_equal(ch0.dobj.ds.basename, ch_pickle.dobj.ds.basename)



# Some tests for the Spherical coordinates handler

import numpy as np
from numpy.testing import assert_almost_equal, assert_equal

from yt.testing import fake_amr_ds

# Our canonical tests are that we can access all of our fields and we can
# compute our volume correctly.


def test_spherical_coordinates():
    # We're going to load up a simple AMR grid and check its volume
    # calculations and path length calculations.
    ds = fake_amr_ds(geometry="spherical")
    axes = ["r", "theta", "phi"]
    for i, axis in enumerate(axes):
        dd = ds.all_data()
        fi = ("index", axis)
        fd = ("index", f"d{axis}")
        ma = np.argmax(dd[fi])
        assert_equal(dd[fi][ma] + dd[fd][ma] / 2.0, ds.domain_right_edge[i].d)
        mi = np.argmin(dd[fi])
        assert_equal(dd[fi][mi] - dd[fd][mi] / 2.0, ds.domain_left_edge[i].d)
        assert_equal(dd[fd].max(), (ds.domain_width / ds.domain_dimensions)[i].d)
    # Note that we're using a lot of funny transforms to get to this, so we do
    # not expect to get actual agreement.  This is a bit of a shame, but I
    # don't think it is avoidable as of right now.  Real datasets will almost
    # certainly be correct, if this is correct to 3 decimel places.
    assert_almost_equal(
        dd[("index", "cell_volume")].sum(dtype="float64"),
        (4.0 / 3.0) * np.pi * ds.domain_width[0] ** 3,
    )
    assert_equal(dd["index", "path_element_r"], dd["index", "dr"])
    assert_equal(
        dd["index", "path_element_theta"], dd["index", "r"] * dd["index", "dtheta"]
    )
    assert_equal(
        dd["index", "path_element_phi"],
        (dd["index", "r"] * dd["index", "dphi"] * np.sin(dd["index", "theta"])),
    )



import pytest

from yt.utilities.on_demand_imports import OnDemand, safe_import


def test_access_available_module():
    class os_imports(OnDemand):
        @safe_import
        def path(self):
            from os import path

            return path

    _os = os_imports()

    _os.path.join("eggs", "saussage")


def test_access_unavailable_module():
    class Bacon_imports(OnDemand):
        @safe_import
        def spam(self):
            from Bacon import spam

            return spam

    _bacon = Bacon_imports()
    with pytest.raises(
        ImportError,
        match=r"No module named 'Bacon'",
    ) as excinfo:
        _bacon.spam()

    # yt should add information to the original error message
    # but this done slightly differently in Python>=3.11
    # (using exception notes), so we can't just match the error message
    # directly. Instead this implements a Python-version agnostic check
    # that the user-visible error message is what we expect.
    complete_error_message = excinfo.exconly()
    assert complete_error_message == (
        "ModuleNotFoundError: No module named 'Bacon'\n"
        "Something went wrong while trying to lazy-import Bacon. "
        "Please make sure that Bacon is properly installed.\n"
        "If the problem persists, please file an issue at "
        "https://github.com/yt-project/yt/issues/new"
    )


def test_class_invalidation():
    with pytest.raises(
        TypeError, match="class .*'s name needs to be suffixed '_imports'"
    ):

        class Bacon(OnDemand):
            pass


def test_base_class_instanciation():
    with pytest.raises(
        TypeError, match="The OnDemand base class cannot be instantiated."
    ):
        OnDemand()



import numpy as np
from numpy.testing import assert_equal

from yt.utilities.lib.cykdtree import utils  # type: ignore
from yt.utilities.lib.cykdtree.tests import assert_less_equal, parametrize


def test_max_pts():
    pts = np.arange(5 * 3).reshape((5, 3)).astype("float64")
    out = utils.py_max_pts(pts)
    np.testing.assert_allclose(out, np.max(pts, axis=0))


def test_min_pts():
    pts = np.arange(5 * 3).reshape((5, 3)).astype("float64")
    out = utils.py_min_pts(pts)
    np.testing.assert_allclose(out, np.min(pts, axis=0))


@parametrize(N=(10), ndim=(2, 3), Lidx=(0, 5), Ridx=(5, 9))
def test_argmax_pts_dim(N=10, ndim=2, Lidx=0, Ridx=9):
    d = ndim - 1
    pts = np.random.rand(N, ndim).astype("float64")
    idx = np.argsort(pts[:, d]).astype("uint64")
    out = utils.py_argmax_pts_dim(pts, idx, d, Lidx, Ridx)
    assert_equal(out, np.argmax(pts[idx[Lidx : (Ridx + 1)], d]) + Lidx)


@parametrize(N=(10), ndim=(2, 3), Lidx=(0, 5), Ridx=(5, 9))
def test_argmin_pts_dim(N=10, ndim=2, Lidx=0, Ridx=9):
    d = ndim - 1
    pts = np.random.rand(N, ndim).astype("float64")
    idx = np.argsort(pts[:, d]).astype("uint64")
    out = utils.py_argmin_pts_dim(pts, idx, d, Lidx, Ridx)
    assert_equal(out, np.argmin(pts[idx[Lidx : (Ridx + 1)], d]) + Lidx)


@parametrize(N=(0, 10, 11), ndim=(2, 3))
def test_quickSort(N=10, ndim=2):
    d = ndim - 1
    np.random.seed(10)
    pts = np.random.rand(N, ndim).astype("float64")
    idx = utils.py_quickSort(pts, d)
    assert_equal(idx.size, N)
    if N != 0:
        np.testing.assert_allclose(idx, np.argsort(pts[:, d]))


@parametrize(N=(0, 10, 11), ndim=(2, 3))
def test_insertSort(N=10, ndim=2):
    d = ndim - 1
    np.random.seed(10)
    pts = np.random.rand(N, ndim).astype("float64")
    idx = utils.py_insertSort(pts, d)
    assert_equal(idx.size, N)
    if N != 0:
        np.testing.assert_allclose(idx, np.argsort(pts[:, d]))


@parametrize(N=(0, 10, 11), ndim=(2, 3))
def test_pivot(N=10, ndim=2):
    d = ndim - 1
    np.random.seed(10)
    pts = np.random.rand(N, ndim).astype("float64")
    q, idx = utils.py_pivot(pts, d)
    if N == 0:
        np.testing.assert_equal(q, -1)
    else:
        piv = pts[idx[q], d]
        nmax = 7 * N / 10 + 6
        assert_less_equal(np.sum(pts[:, d] < piv), nmax)
        assert_less_equal(np.sum(pts[:, d] > piv), nmax)


@parametrize(N=(0, 10, 11), ndim=(2, 3))
def test_partition(N=10, ndim=2):
    d = ndim - 1
    p = 0
    np.random.seed(10)
    pts = np.random.rand(N, ndim).astype("float64")
    q, idx = utils.py_partition(pts, d, p)
    if N == 0:
        assert_equal(q, -1)
    else:
        piv = pts[p, d]
        np.testing.assert_approx_equal(pts[idx[q], d], piv)
        np.testing.assert_array_less(pts[idx[:q], d], piv)
        np.testing.assert_array_less(piv, pts[idx[(q + 1) :], d])


@parametrize(N=(0, 10, 11), ndim=(2, 3))
def test_partition_given_pivot(N=10, ndim=2):
    d = ndim - 1
    np.random.seed(10)
    pts = np.random.rand(N, ndim).astype("float64")
    if N == 0:
        piv_list = [0.5]
    else:
        piv_list = [0.5, np.median(pts[:, d])]
    for piv in piv_list:
        q, idx = utils.py_partition_given_pivot(pts, d, piv)
        if N == 0:
            assert_equal(q, -1)
        else:
            assert_less_equal(pts[idx[q], d], piv)
            np.testing.assert_array_less(pts[idx[:q], d], piv)
            np.testing.assert_array_less(piv, pts[idx[(q + 1) :], d])


@parametrize(N=(0, 10, 11), ndim=(2, 3))
def test_select(N=10, ndim=2):
    d = ndim - 1
    np.random.seed(10)
    pts = np.random.rand(N, ndim).astype("float64")
    p = int(N) // 2 + int(N) % 2
    q, idx = utils.py_select(pts, d, p)
    assert_equal(idx.size, N)
    if N == 0:
        assert_equal(q, -1)
    else:
        assert_equal(q, p - 1)
        med = np.median(pts[:, d])
        np.testing.assert_array_less(pts[idx[:q], d], med)
        np.testing.assert_array_less(med, pts[idx[(q + 1) :], d])
        if N % 2:
            np.testing.assert_approx_equal(pts[idx[q], d], med)
        else:
            np.testing.assert_array_less(pts[idx[q], d], med)


@parametrize(N=(0, 10, 11), ndim=(2, 3), use_sliding_midpoint=(False, True))
def test_split(N=10, ndim=2, use_sliding_midpoint=False):
    np.random.seed(10)
    pts = np.random.rand(N, ndim).astype("float64")
    p = int(N) // 2 + int(N) % 2
    q, d, idx = utils.py_split(pts, use_sliding_midpoint=use_sliding_midpoint)
    assert_equal(idx.size, N)
    if N == 0:
        assert_equal(q, -1)
    else:
        if use_sliding_midpoint:
            # Midpoint
            med = 0.5 * (np.min(pts[:, d]) + np.max(pts[:, d]))
            np.testing.assert_array_less(pts[idx[:q], d], med)
            np.testing.assert_array_less(med, pts[idx[(q + 1) :], d])
            np.testing.assert_array_less(pts[idx[q], d], med)
            # Sliding midpoint (slide to minimum)
            q, d, idx = utils.py_split(
                pts,
                mins=-1 * np.ones(ndim),
                maxs=np.ones(ndim),
                use_sliding_midpoint=True,
            )
            med = np.min(pts[:, d])
            assert_equal(q, 0)
            np.testing.assert_array_less(pts[idx[:q], d], med)
            np.testing.assert_array_less(med, pts[idx[(q + 1) :], d])
            np.testing.assert_approx_equal(pts[idx[q], d], med)
            # Sliding midpoint (slide to maximum)
            q, d, idx = utils.py_split(
                pts,
                mins=np.zeros(ndim),
                maxs=2 * np.ones(ndim),
                use_sliding_midpoint=True,
            )
            med = np.max(pts[:, d])
            assert_equal(q, N - 2)
            np.testing.assert_array_less(pts[idx[: (q + 1)], d], med)
            np.testing.assert_approx_equal(pts[idx[q + 1], d], med)
        else:
            assert_equal(q, p - 1)
            med = np.median(pts[:, d])
            np.testing.assert_array_less(pts[idx[:q], d], med)
            np.testing.assert_array_less(med, pts[idx[(q + 1) :], d])
            if N % 2:
                np.testing.assert_approx_equal(pts[idx[q], d], med)
            else:
                np.testing.assert_array_less(pts[idx[q], d], med)



import tempfile
import time

import numpy as np
from nose.tools import assert_raises

import yt.utilities.lib.cykdtree as cykdtree
from yt.utilities.lib.cykdtree.tests import (
    make_points,
    make_points_neighbors,
    parametrize,
)


@parametrize(
    npts=100, ndim=(2, 3), periodic=(False, True), use_sliding_midpoint=(False, True)
)
def test_PyKDTree(npts=100, ndim=2, periodic=False, use_sliding_midpoint=False):
    pts, le, re, ls = make_points(npts, ndim)
    cykdtree.PyKDTree(
        pts,
        le,
        re,
        leafsize=ls,
        periodic=periodic,
        use_sliding_midpoint=use_sliding_midpoint,
    )


def test_PyKDTree_errors():
    pts, le, re, ls = make_points(100, 2)
    assert_raises(ValueError, cykdtree.PyKDTree, pts, le, re, leafsize=1)


@parametrize(npts=100, ndim=(2, 3), periodic=(False, True))
def test_search(npts=100, ndim=2, periodic=False):
    pts, le, re, ls = make_points(npts, ndim)
    tree = cykdtree.PyKDTree(pts, le, re, leafsize=ls, periodic=periodic)
    pos_list = [le, (le + re) / 2.0]
    if periodic:
        pos_list.append(re)
    for pos in pos_list:
        leaf = tree.get(pos)
        leaf.neighbors


@parametrize(npts=100, ndim=(2, 3))
def test_search_errors(npts=100, ndim=2):
    pts, le, re, ls = make_points(npts, ndim)
    tree = cykdtree.PyKDTree(pts, le, re, leafsize=ls)
    assert_raises(ValueError, tree.get, re)


@parametrize(periodic=(False, True))
def test_neighbors(periodic=False):
    pts, le, re, ls, left_neighbors, right_neighbors = make_points_neighbors(
        periodic=periodic
    )
    tree = cykdtree.PyKDTree(pts, le, re, leafsize=ls, periodic=periodic)
    for leaf in tree.leaves:
        out_str = str(leaf.id)
        try:
            for d in range(tree.ndim):
                out_str += "\nleft:  {} {} {}".format(
                    d, leaf.left_neighbors[d], left_neighbors[d][leaf.id]
                )
                assert len(left_neighbors[d][leaf.id]) == len(leaf.left_neighbors[d])
                for i in range(len(leaf.left_neighbors[d])):
                    assert left_neighbors[d][leaf.id][i] == leaf.left_neighbors[d][i]
                out_str += "\nright: {} {} {}".format(
                    d, leaf.right_neighbors[d], right_neighbors[d][leaf.id]
                )
                assert len(right_neighbors[d][leaf.id]) == len(leaf.right_neighbors[d])
                for i in range(len(leaf.right_neighbors[d])):
                    assert right_neighbors[d][leaf.id][i] == leaf.right_neighbors[d][i]
        except Exception as e:
            for leaf in tree.leaves:
                print(leaf.id, leaf.left_edge, leaf.right_edge)
            print(out_str)
            raise e


@parametrize(npts=100, ndim=(2, 3), periodic=(False, True))
def test_get_neighbor_ids(npts=100, ndim=2, periodic=False):
    pts, le, re, ls = make_points(npts, ndim)
    tree = cykdtree.PyKDTree(pts, le, re, leafsize=ls, periodic=periodic)
    pos_list = [le, (le + re) / 2.0]
    if periodic:
        pos_list.append(re)
    for pos in pos_list:
        tree.get_neighbor_ids(pos)


def time_tree_construction(Ntime, LStime, ndim=2):
    pts, le, re, ls = make_points(Ntime, ndim, leafsize=LStime)
    t0 = time.time()
    cykdtree.PyKDTree(pts, le, re, leafsize=LStime)
    t1 = time.time()
    print(f"{Ntime} {ndim}D points, leafsize {LStime}: took {t1 - t0} s")


def time_neighbor_search(Ntime, LStime, ndim=2):
    pts, le, re, ls = make_points(Ntime, ndim, leafsize=LStime)
    tree = cykdtree.PyKDTree(pts, le, re, leafsize=LStime)
    t0 = time.time()
    tree.get_neighbor_ids(0.5 * np.ones(tree.ndim, "double"))
    t1 = time.time()
    print(f"{Ntime} {ndim}D points, leafsize {LStime}: took {t1 - t0} s")


def test_save_load():
    for periodic in (True, False):
        for ndim in range(1, 5):
            pts, le, re, ls = make_points(100, ndim)
            tree = cykdtree.PyKDTree(
                pts, le, re, leafsize=ls, periodic=periodic, data_version=ndim + 12
            )
            with tempfile.NamedTemporaryFile(delete=False) as tf:
                tree.save(tf.name)
                restore_tree = cykdtree.PyKDTree.from_file(tf.name)
                tree.assert_equal(restore_tree)



import glob
import importlib
import os
import sys
import time
import types

_frontends = [
    "adaptahop",
    "ahf",
    "amrvac",
    "art",
    "arepo",
    "artio",
    "athena",
    "athena_pp",
    "boxlib",
    "cf_radial",
    "chimera",
    "chombo",
    "cholla",
    "eagle",
    "enzo_e",
    "enzo",
    "exodus_ii",
    "fits",
    "flash",
    "gadget",
    "gadget_fof",
    "gamer",
    "gdf",
    "gizmo",
    "halo_catalog",
    "http_stream",
    "moab",
    "nc4_cm1",
    "open_pmd",
    "owls",
    "owls_subfind",
    "ramses",
    "rockstar",
    "sdf",
    "stream",
    "swift",
    "tipsy",
    "ytdata",
]


class _frontend_container:
    def __init__(self):
        for frontend in _frontends:
            _mod = f"yt.frontends.{frontend}.api"
            setattr(self, frontend, importlib.import_module(_mod))
        setattr(self, "api", importlib.import_module("yt.frontends.api"))
        setattr(self, "__name__", "yt.frontends.api")




cdef class BooleanSelector(SelectorObject):

    def __init__(self, dobj):
        # Note that this has a different API than the other selector objects,
        # so will not work as a traditional data selector.
        if not hasattr(dobj.dobj1, "selector"):
            self.sel1 = dobj.dobj1
        else:
            self.sel1 = dobj.dobj1.selector
        if not hasattr(dobj.dobj2, "selector"):
            self.sel2 = dobj.dobj2
        else:
            self.sel2 = dobj.dobj2.selector

cdef class BooleanANDSelector(BooleanSelector):
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
        if rv2 == 0: return 0
        return 1

    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                              np.float64_t right_edge[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_bbox_edge(left_edge, right_edge)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_bbox_edge(left_edge, right_edge)
        if rv2 == 0: return 0
        return max(rv1, rv2)

    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
        if rv2 == 0: return 0
        return 1

    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_cell(pos, dds)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_cell(pos, dds)
        if rv2 == 0: return 0
        return 1

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_point(pos)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_point(pos)
        if rv2 == 0: return 0
        return 1

    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        cdef int rv1 = self.sel1.select_sphere(pos, radius)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_sphere(pos, radius)
        if rv2 == 0: return 0
        return 1

    def _hash_vals(self):
        return (self.sel1._hash_vals() +
                ("and",) +
                self.sel2._hash_vals())

cdef class BooleanORSelector(BooleanSelector):
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_bbox(left_edge, right_edge)
        if rv1 == 1: return 1
        cdef int rv2 = self.sel2.select_bbox(left_edge, right_edge)
        if rv2 == 1: return 1
        return 0

    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                              np.float64_t right_edge[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_bbox_edge(left_edge, right_edge)
        if rv1 == 1: return 1
        cdef int rv2 = self.sel2.select_bbox_edge(left_edge, right_edge)
        if rv2 == 1: return 1
        return max(rv1, rv2)

    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        cdef int rv1 = self.sel1.select_grid(left_edge, right_edge, level, o)
        if rv1 == 1: return 1
        cdef int rv2 = self.sel2.select_grid(left_edge, right_edge, level, o)
        if rv2 == 1: return 1
        if (rv1 == 2) or (rv2 == 2): return 2
        return 0

    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_cell(pos, dds)
        if rv1 == 1: return 1
        cdef int rv2 = self.sel2.select_cell(pos, dds)
        if rv2 == 1: return 1
        return 0

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_point(pos)
        if rv1 == 1: return 1
        cdef int rv2 = self.sel2.select_point(pos)
        if rv2 == 1: return 1
        return 0

    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        cdef int rv1 = self.sel1.select_sphere(pos, radius)
        if rv1 == 1: return 1
        cdef int rv2 = self.sel2.select_sphere(pos, radius)
        if rv2 == 1: return 1
        return 0

    def _hash_vals(self):
        return (self.sel1._hash_vals() +
                ("or",) +
                self.sel2._hash_vals())

cdef class BooleanNOTSelector(BooleanSelector):
    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        # We always return True here, because we don't have a "fully included"
        # check anywhere else.
        return 1

    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                              np.float64_t right_edge[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_bbox_edge(left_edge, right_edge)
        if rv1 == 0: return 1
        elif rv1 == 1: return 0
        return 2

    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        return 1

    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_cell(pos, dds)
        if rv1 == 0: return 1
        return 0

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_point(pos)
        if rv1 == 0: return 1
        return 0

    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        cdef int rv1 = self.sel1.select_sphere(pos, radius)
        if rv1 == 0: return 1
        return 0

    def _hash_vals(self):
        return (self.sel1._hash_vals() +
                ("not",))

cdef class BooleanXORSelector(BooleanSelector):

    cdef int select_bbox(self, np.float64_t left_edge[3],
                               np.float64_t right_edge[3]) noexcept nogil:
        # We always return True here, because we don't have a "fully included"
        # check anywhere else.
        return 1

    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                              np.float64_t right_edge[3]) noexcept nogil:
        # Return 2 in cases where one or both selectors partially overlap since
        # part of the bounding box could satisfy the condition unless the
        # selectors are identical.
        cdef int rv1 = self.sel1.select_bbox_edge(left_edge, right_edge)
        cdef int rv2 = self.sel2.select_bbox_edge(left_edge, right_edge)
        if rv1 == rv2:
            if rv1 == 2:
                # If not identical, part of the bbox will be touched by one
                # selector and not the other.
                # if self.sel1 == self.sel2: return 0  # requires gil
                return 2
            return 0
        if rv1 == 0: return rv2
        if rv2 == 0: return rv1
        return 2  # part of bbox only touched by selector fully covering bbox

    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        return 1

    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_cell(pos, dds)
        cdef int rv2 = self.sel2.select_cell(pos, dds)
        if rv1 == rv2: return 0
        return 1

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_point(pos)
        cdef int rv2 = self.sel2.select_point(pos)
        if rv1 == rv2: return 0
        return 1

    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        cdef int rv1 = self.sel1.select_sphere(pos, radius)
        cdef int rv2 = self.sel2.select_sphere(pos, radius)
        if rv1 == rv2: return 0
        return 1

    def _hash_vals(self):
        return (self.sel1._hash_vals() +
                ("xor",) +
                self.sel2._hash_vals())

cdef class BooleanNEGSelector(BooleanSelector):

    cdef int select_bbox(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3]) noexcept nogil:
        # We always return True here, because we don't have a "fully included"
        # check anywhere else.
        return self.sel1.select_bbox(left_edge, right_edge)

    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                              np.float64_t right_edge[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_bbox_edge(left_edge, right_edge)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_bbox_edge(left_edge, right_edge)
        if rv2 == 1:
            return 0
        elif rv2 == 0:
            return rv1
        # If sel2 is partial, then sel1 - sel2 will be partial as long
        # as sel1 != sel2
        # if self.sel1 == self.sel2: return 0  # requires gil
        return 2

    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        return self.sel1.select_grid(left_edge, right_edge, level, o)

    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_cell(pos, dds)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_cell(pos, dds)
        if rv2 == 1: return 0
        return 1

    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        cdef int rv1 = self.sel1.select_point(pos)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_point(pos)
        if rv2 == 1: return 0
        return 1

    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        cdef int rv1 = self.sel1.select_sphere(pos, radius)
        if rv1 == 0: return 0
        cdef int rv2 = self.sel2.select_sphere(pos, radius)
        if rv2 == 1: return 0
        return 1

    def _hash_vals(self):
        return (self.sel1._hash_vals() +
                ("neg",) +
                self.sel2._hash_vals())

cdef class ChainedBooleanSelector(SelectorObject):
    cdef int n_obj
    cdef np.ndarray selectors
    def __init__(self, dobj):
        # These are data objects, not selectors
        self.n_obj = len(dobj.data_objects)
        self.selectors = np.empty(self.n_obj, dtype="object")
        for i in range(self.n_obj):
            self.selectors[i] = dobj.data_objects[i].selector

cdef class ChainedBooleanANDSelector(ChainedBooleanSelector):
    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3]) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_bbox(
                        left_edge, right_edge) == 0:
                    return 0
        return 1

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                              np.float64_t right_edge[3]) noexcept nogil:
        cdef int selected = 1
        cdef int ret
        with gil:
            for i in range(self.n_obj):
                ret = (<SelectorObject>self.selectors[i]).select_bbox_edge(
                    left_edge, right_edge)
                if ret == 0:
                    return 0
                elif ret == 2:
                    selected = 2
        return selected

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_grid(
                        left_edge, right_edge, level, o) == 0:
                    return 0
        return 1

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_cell(
                        pos, dds) == 0:
                    return 0
        return 1

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_point(pos) == 0:
                    return 0
        return 1

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_sphere(
                        pos, radius) == 0:
                    return 0
        return 1

    def _hash_vals(self):
        v = ("chained_and",)
        for s in self.selectors:
            v += s._hash_vals()
        return v

intersection_selector = ChainedBooleanANDSelector

cdef class ChainedBooleanORSelector(ChainedBooleanSelector):
    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_bbox(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3]) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_bbox(
                        left_edge, right_edge) == 1:
                    return 1
        return 0

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_bbox_edge(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3]) noexcept nogil:
        cdef int selected = 0
        cdef int ret
        with gil:
            for i in range(self.n_obj):
                ret = (<SelectorObject>self.selectors[i]).select_bbox_edge(
                    left_edge, right_edge)
                if ret == 1:
                    return 1
                elif ret == 2:
                    selected = 2
        return selected

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_grid(self, np.float64_t left_edge[3],
                         np.float64_t right_edge[3], np.int32_t level,
                         Oct *o = NULL) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_grid(
                        left_edge, right_edge, level, o) == 1:
                    return 1
        return 0

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_cell(self, np.float64_t pos[3], np.float64_t dds[3]) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_cell(
                        pos, dds) == 1:
                    return 1
        return 0

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_point(self, np.float64_t pos[3]) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_point(pos) == 1:
                    return 1
        return 0

    @cython.cdivision(True)
    @cython.boundscheck(False)
    @cython.wraparound(False)
    cdef int select_sphere(self, np.float64_t pos[3], np.float64_t radius) noexcept nogil:
        with gil:
            for i in range(self.n_obj):
                if (<SelectorObject>self.selectors[i]).select_sphere(
                        pos, radius) == 1:
                    return 1
        return 0

    def _hash_vals(self):
        v = ("chained_or",)
        for s in self.selectors:
            v += s._hash_vals()
        return v

union_selector = ChainedBooleanORSelector



import numpy as np

import yt

# Load the dataset.
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# Choose a center for the render.
c = [0.5, 0.5, 0.5]

# Our image plane will be normal to some vector.  For things like collapsing
# objects, you could set it the way you would a cutting plane -- but for this
# dataset, we'll just choose an off-axis value at random.  This gets normalized
# automatically.
L = [1.0, 0.0, 0.0]

# Our "width" is the width of the image plane as well as the depth.
# The first element is the left to right width, the second is the
# top-bottom width, and the last element is the back-to-front width
# (all in code units)
W = [0.04, 0.04, 0.4]

# The number of pixels along one side of the image.
# The final image will have Npixel^2 pixels.
Npixels = 512

# Create the off axis projection.
# Setting no_ghost to False speeds up the process, but makes a
# slightly lower quality image.
image = yt.off_axis_projection(ds, c, L, W, Npixels, ("gas", "density"), no_ghost=False)

# Write out the final image and give it a name
# relating to what our dataset is called.
# We save the log of the values so that the colors do not span
# many orders of magnitude.  Try it without and see what happens.
yt.write_image(np.log10(image), f"{ds}_offaxis_projection.png")



"""
Oct definitions file




"""


cimport cython
cimport numpy as np
from libc.math cimport floor
from libc.stdlib cimport bsearch, free, malloc, qsort, realloc

from yt.geometry cimport oct_visitors, selection_routines
from yt.utilities.lib.allocation_container cimport AllocationContainer, ObjectPool
from yt.utilities.lib.fp_utils cimport *

from .oct_visitors cimport Oct, OctInfo, OctVisitor, cind


cdef int ORDER_MAX

cdef struct OctKey:
    np.int64_t key
    Oct *node
    # These next two are for particle sparse octrees.
    np.int64_t *indices
    np.int64_t pcount

cdef struct OctList

cdef struct OctList:
    OctList *next
    Oct *o

# NOTE: This object *has* to be the same size as the AllocationContainer
# object.  There's an assert in the __cinit__ function.
cdef struct OctAllocationContainer:
    np.uint64_t n
    np.uint64_t n_assigned
    np.uint64_t offset
    np.int64_t con_id # container id
    Oct *my_objs

cdef class OctObjectPool(ObjectPool):
    cdef inline OctAllocationContainer *get_cont(self, int i):
        return <OctAllocationContainer*> (&self.containers[i])

cdef OctList *OctList_append(OctList *list, Oct *o)
cdef int OctList_count(OctList *list)
cdef void OctList_delete(OctList *list)

cdef class OctreeContainer:
    cdef public OctObjectPool domains
    cdef Oct ****root_mesh
    cdef int partial_coverage
    cdef int level_offset
    cdef int nn[3]
    cdef np.uint8_t nz
    cdef np.float64_t DLE[3]
    cdef np.float64_t DRE[3]
    cdef public np.int64_t nocts
    cdef public int num_domains
    cdef Oct *get(self, np.float64_t ppos[3], OctInfo *oinfo = ?,
                  int max_level = ?) nogil
    cdef int get_root(self, int ind[3], Oct **o) nogil
    cdef Oct **neighbors(self, OctInfo *oinfo, np.int64_t *nneighbors,
                         Oct *o, bint periodicity[3])
    # This function must return the offset from global-to-local domains; i.e.,
    # AllocationContainer.offset if such a thing exists.
    cdef np.int64_t get_domain_offset(self, int domain_id)
    cdef void visit_all_octs(self,
                        selection_routines.SelectorObject selector,
                        OctVisitor visitor,
                        int vc = ?, np.int64_t *indices = ?)
    cdef Oct *next_root(self, int domain_id, int ind[3])
    cdef Oct *next_child(self, int domain_id, int ind[3], Oct *parent) except? NULL
    cdef void append_domain(self, np.int64_t domain_count)
    # The fill_style is the ordering, C or F, of the octs in the file.  "o"
    # corresponds to C, and "r" is for Fortran.
    cdef public object fill_style

cdef class SparseOctreeContainer(OctreeContainer):
    cdef OctKey *root_nodes
    cdef void *tree_root
    cdef int num_root
    cdef int max_root
    cdef void key_to_ipos(self, np.int64_t key, np.int64_t pos[3])
    cdef np.int64_t ipos_to_key(self, int pos[3]) nogil

cdef class RAMSESOctreeContainer(SparseOctreeContainer):
    pass

cdef extern from "tsearch.h" nogil:
    void *tsearch(const void *key, void **rootp,
                    int (*compar)(const void *, const void *))
    void *tfind(const void *key, const void **rootp,
                    int (*compar)(const void *, const void *))
    void *tdelete(const void *key, void **rootp,
                    int (*compar)(const void *, const void *))



import os
import shutil
import tempfile

import numpy as np
from numpy.testing import assert_array_equal, assert_equal

from yt.data_objects.api import create_profile
from yt.frontends.ytdata.api import (
    YTDataContainerDataset,
    YTGridDataset,
    YTNonspatialDataset,
    YTProfileDataset,
    YTSpatialPlotDataset,
    save_as_dataset,
)
from yt.loaders import load
from yt.testing import assert_allclose_units
from yt.units.yt_array import YTArray, YTQuantity
from yt.utilities.answer_testing.framework import (
    AnswerTestingTest,
    data_dir_load,
    requires_ds,
)
from yt.visualization.profile_plotter import PhasePlot, ProfilePlot


def make_tempdir():
    if int(os.environ.get("GENERATE_YTDATA", 0)):
        return "."
    else:
        return tempfile.mkdtemp()


def compare_unit_attributes(ds1, ds2):
    r"""
    Checks to make sure that the length, mass, time, velocity, and
    magnetic units are the same for two different dataset objects.
    """
    attrs = ("length_unit", "mass_unit", "time_unit", "velocity_unit", "magnetic_unit")
    for attr in attrs:
        u1 = getattr(ds1, attr, None)
        u2 = getattr(ds2, attr, None)
        assert u1 == u2


class YTDataFieldTest(AnswerTestingTest):
    _type_name = "YTDataTest"
    _attrs = ("field_name",)

    def __init__(self, ds_fn, field, decimals=10, geometric=True):
        super().__init__(ds_fn)
        self.field = field
        if isinstance(field, tuple) and len(field) == 2:
            self.field_name = field[1]
        else:
            self.field_name = field
        self.decimals = decimals
        self.geometric = geometric

    def run(self):
        if self.geometric:
            obj = self.ds.all_data()
        else:
            obj = self.ds.data
        num_e = obj[self.field].size
        avg = obj[self.field].mean()
        return np.array([num_e, avg])

    def compare(self, new_result, old_result):
        err_msg = f"YTData field values for {self.field} not equal."
        if self.decimals is None:
            assert_equal(new_result, old_result, err_msg=err_msg, verbose=True)
        else:
            assert_allclose_units(
                new_result,
                old_result,
                10.0 ** (-self.decimals),
                err_msg=err_msg,
                verbose=True,
            )


enzotiny = "enzo_tiny_cosmology/DD0046/DD0046"


@requires_ds(enzotiny)
def test_datacontainer_data():
    tmpdir = make_tempdir()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    ds = data_dir_load(enzotiny)
    sphere = ds.sphere(ds.domain_center, (10, "Mpc"))
    fn = sphere.save_as_dataset(fields=[("gas", "density"), ("all", "particle_mass")])
    full_fn = os.path.join(tmpdir, fn)
    sphere_ds = load(full_fn)
    compare_unit_attributes(ds, sphere_ds)
    assert isinstance(sphere_ds, YTDataContainerDataset)
    yield YTDataFieldTest(full_fn, ("grid", "density"))
    yield YTDataFieldTest(full_fn, ("all", "particle_mass"))
    cr = ds.cut_region(sphere, ['obj[("gas", "temperature")] > 1e4'])
    fn = cr.save_as_dataset(fields=[("gas", "temperature")])
    full_fn = os.path.join(tmpdir, fn)
    cr_ds = load(full_fn)
    assert isinstance(cr_ds, YTDataContainerDataset)
    assert (cr[("gas", "temperature")] == cr_ds.data[("gas", "temperature")]).all()
    os.chdir(curdir)
    if tmpdir != ".":
        shutil.rmtree(tmpdir)


@requires_ds(enzotiny)
def test_grid_datacontainer_data():
    tmpdir = make_tempdir()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    ds = data_dir_load(enzotiny)

    cg = ds.covering_grid(level=0, left_edge=[0.25] * 3, dims=[16] * 3)
    fn = cg.save_as_dataset(
        fields=[
            ("gas", "density"),
            ("all", "particle_mass"),
            ("all", "particle_position"),
        ]
    )
    full_fn = os.path.join(tmpdir, fn)
    cg_ds = load(full_fn)
    compare_unit_attributes(ds, cg_ds)
    assert isinstance(cg_ds, YTGridDataset)
    assert (
        cg["all", "particle_position"].shape
        == cg_ds.r["all", "particle_position"].shape
    )
    yield YTDataFieldTest(full_fn, ("grid", "density"))
    yield YTDataFieldTest(full_fn, ("all", "particle_mass"))

    ag = ds.arbitrary_grid(left_edge=[0.25] * 3, right_edge=[0.75] * 3, dims=[16] * 3)
    fn = ag.save_as_dataset(fields=[("gas", "density"), ("all", "particle_mass")])
    full_fn = os.path.join(tmpdir, fn)
    ag_ds = load(full_fn)
    compare_unit_attributes(ds, ag_ds)
    assert isinstance(ag_ds, YTGridDataset)
    yield YTDataFieldTest(full_fn, ("grid", "density"))
    yield YTDataFieldTest(full_fn, ("all", "particle_mass"))

    my_proj = ds.proj(("gas", "density"), "x", weight_field=("gas", "density"))
    frb = my_proj.to_frb(1.0, (800, 800))
    fn = frb.save_as_dataset(fields=[("gas", "density")])
    frb_ds = load(fn)
    assert_array_equal(frb[("gas", "density")], frb_ds.data[("gas", "density")])
    compare_unit_attributes(ds, frb_ds)
    assert isinstance(frb_ds, YTGridDataset)
    yield YTDataFieldTest(full_fn, ("grid", "density"), geometric=False)
    os.chdir(curdir)
    if tmpdir != ".":
        shutil.rmtree(tmpdir)


@requires_ds(enzotiny)
def test_spatial_data():
    tmpdir = make_tempdir()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    ds = data_dir_load(enzotiny)
    proj = ds.proj(("gas", "density"), "x", weight_field=("gas", "density"))
    fn = proj.save_as_dataset()
    full_fn = os.path.join(tmpdir, fn)
    proj_ds = load(full_fn)
    compare_unit_attributes(ds, proj_ds)
    assert isinstance(proj_ds, YTSpatialPlotDataset)
    yield YTDataFieldTest(full_fn, ("grid", "density"), geometric=False)
    os.chdir(curdir)
    if tmpdir != ".":
        shutil.rmtree(tmpdir)


@requires_ds(enzotiny)
def test_profile_data():
    tmpdir = make_tempdir()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    ds = data_dir_load(enzotiny)
    ad = ds.all_data()
    profile_1d = create_profile(
        ad,
        ("gas", "density"),
        ("gas", "temperature"),
        weight_field=("gas", "cell_mass"),
    )
    fn = profile_1d.save_as_dataset()
    full_fn = os.path.join(tmpdir, fn)
    prof_1d_ds = load(full_fn)
    compare_unit_attributes(ds, prof_1d_ds)
    assert isinstance(prof_1d_ds, YTProfileDataset)

    for field in profile_1d.standard_deviation:
        assert_array_equal(
            profile_1d.standard_deviation[field],
            prof_1d_ds.profile.standard_deviation["data", field[1]],
        )

    p1 = ProfilePlot(
        prof_1d_ds.data,
        ("gas", "density"),
        ("gas", "temperature"),
        weight_field=("gas", "cell_mass"),
    )
    p1.save()

    yield YTDataFieldTest(full_fn, ("data", "temperature"), geometric=False)
    yield YTDataFieldTest(full_fn, ("data", "x"), geometric=False)
    yield YTDataFieldTest(full_fn, ("data", "density"), geometric=False)
    profile_2d = create_profile(
        ad,
        [("gas", "density"), ("gas", "temperature")],
        ("gas", "cell_mass"),
        weight_field=None,
        n_bins=(128, 128),
    )
    fn = profile_2d.save_as_dataset()
    full_fn = os.path.join(tmpdir, fn)
    prof_2d_ds = load(full_fn)
    compare_unit_attributes(ds, prof_2d_ds)
    assert isinstance(prof_2d_ds, YTProfileDataset)

    p2 = PhasePlot(
        prof_2d_ds.data,
        ("gas", "density"),
        ("gas", "temperature"),
        ("gas", "cell_mass"),
        weight_field=None,
    )
    p2.save()

    yield YTDataFieldTest(full_fn, ("data", "density"), geometric=False)
    yield YTDataFieldTest(full_fn, ("data", "x"), geometric=False)
    yield YTDataFieldTest(full_fn, ("data", "temperature"), geometric=False)
    yield YTDataFieldTest(full_fn, ("data", "y"), geometric=False)
    yield YTDataFieldTest(full_fn, ("data", "cell_mass"), geometric=False)
    os.chdir(curdir)
    if tmpdir != ".":
        shutil.rmtree(tmpdir)


@requires_ds(enzotiny)
def test_nonspatial_data():
    tmpdir = make_tempdir()
    curdir = os.getcwd()
    os.chdir(tmpdir)
    ds = data_dir_load(enzotiny)
    region = ds.box([0.25] * 3, [0.75] * 3)
    sphere = ds.sphere(ds.domain_center, (10, "Mpc"))
    my_data = {}
    my_data["region_density"] = region[("gas", "density")]
    my_data["sphere_density"] = sphere[("gas", "density")]
    fn = "test_data.h5"
    save_as_dataset(ds, fn, my_data)
    full_fn = os.path.join(tmpdir, fn)
    array_ds = load(full_fn)
    compare_unit_attributes(ds, array_ds)
    assert isinstance(array_ds, YTNonspatialDataset)
    yield YTDataFieldTest(full_fn, "region_density", geometric=False)
    yield YTDataFieldTest(full_fn, "sphere_density", geometric=False)

    my_data = {"density": YTArray(np.linspace(1.0, 20.0, 10), "g/cm**3")}
    fake_ds = {"current_time": YTQuantity(10, "Myr")}
    fn = "random_data.h5"
    save_as_dataset(fake_ds, fn, my_data)
    full_fn = os.path.join(tmpdir, fn)
    new_ds = load(full_fn)
    assert isinstance(new_ds, YTNonspatialDataset)
    yield YTDataFieldTest(full_fn, ("data", "density"), geometric=False)
    os.chdir(curdir)
    if tmpdir != ".":
        shutil.rmtree(tmpdir)



import pytest

import yt
from yt.testing import requires_file
from yt.utilities.answer_testing.framework import data_dir_load_v2


class TestRawFieldSlice:
    @requires_file("Laser/plt00015")
    @classmethod
    def setup_class(cls):
        cls.ds = data_dir_load_v2("Laser/plt00015")
        fields = [
            ("raw", "Bx"),
            ("raw", "By"),
            ("raw", "Bz"),
            ("raw", "Ex"),
            ("raw", "Ey"),
            ("raw", "Ez"),
            ("raw", "jx"),
            ("raw", "jy"),
            ("raw", "jz"),
        ]
        cls.sl = yt.SlicePlot(cls.ds, "z", fields)
        cls.sl.set_log("all", False)
        cls.sl.render()

    @pytest.mark.parametrize(
        "fname", ["Bx", "By", "Bz", "Ex", "Ey", "Ez", "jx", "jy", "jz"]
    )
    @pytest.mark.mpl_image_compare
    def test_raw_field_slice(self, fname):
        return self.sl.plots["raw", "Bx"].figure



from numpy.testing import assert_equal

from yt.frontends.gdf.api import GDFDataset
from yt.testing import requires_file, requires_module, units_override_check
from yt.utilities.answer_testing.framework import (
    data_dir_load,
    requires_ds,
    small_patch_amr,
)

_fields = [("gas", "density"), ("gas", "velocity_x")]

sedov = "sedov/sedov_tst_0004.h5"


@requires_ds(sedov)
def test_sedov_tunnel():
    ds = data_dir_load(sedov)
    assert_equal(str(ds), "sedov_tst_0004")
    for test in small_patch_amr(ds, _fields):
        test_sedov_tunnel.__name__ = test.description
        yield test


@requires_module("h5py")
@requires_file(sedov)
def test_GDFDataset():
    assert isinstance(data_dir_load(sedov), GDFDataset)


@requires_module("h5py")
@requires_file(sedov)
def test_units_override():
    units_override_check(sedov)



import yt

# load the dataset
ds = yt.load("IsolatedGalaxy/galaxy0030/galaxy0030")

# This is an object that describes the entire box
ad = ds.all_data()

# We plot the average velocity magnitude (mass-weighted) in our object
# as a function of density and temperature
plot = yt.PhasePlot(
    ad, ("gas", "density"), ("gas", "temperature"), ("gas", "velocity_magnitude")
)

# save the plot
plot.save()



from numpy.testing import assert_raises

from yt.utilities.logger import set_log_level


def test_valid_level():
    # test a subset of valid entries to cover
    # - case-insensitivity
    # - integer values
    # - "all" alias, which isn't standard
    for lvl in ("all", "ALL", 10, 42, "info", "warning", "ERROR", "CRITICAL"):
        set_log_level(lvl)


def test_invalid_level():
    # these are the exceptions raised by logging.Logger.setLog
    # since they are perfectly clear and readable, we check that nothing else
    # happens in the wrapper
    assert_raises(TypeError, set_log_level, 1.5)
    assert_raises(ValueError, set_log_level, "invalid_level")



import numpy as np
from numpy.testing import assert_array_less, assert_equal

from yt.testing import fake_random_ds
from yt.utilities.math_utils import periodic_dist


def setup():
    from yt.config import ytcfg

    ytcfg["yt", "internals", "within_testing"] = True


def test_point_selector():
    # generate fake amr data
    bbox = np.array([[-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0]])
    ds = fake_random_ds(16, nprocs=7, bbox=bbox)
    assert all(ds.periodicity)

    dd = ds.all_data()
    positions = np.array([dd[("index", ax)] for ax in "xyz"]).T
    delta = 0.5 * np.array([dd[("index", f"d{ax}")] for ax in "xyz"]).T
    # ensure cell centers and corners always return one and
    # only one point object
    for p in positions:
        data = ds.point(p)
        assert_equal(data[("index", "ones")].shape[0], 1)
    for p in positions - delta:
        data = ds.point(p)
        assert_equal(data[("index", "ones")].shape[0], 1)
    for p in positions + delta:
        data = ds.point(p)
        assert_equal(data[("index", "ones")].shape[0], 1)


def test_sphere_selector():
    # generate fake data with a number of non-cubical grids
    ds = fake_random_ds(64, nprocs=51)
    assert all(ds.periodicity)

    # aligned tests
    spheres = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5], [1.0, 1.0, 1.0], [0.25, 0.75, 0.25]]

    for center in spheres:
        data = ds.sphere(center, 0.25)
        # WARNING: this value has not be externally verified
        dd = ds.all_data()
        dd.set_field_parameter("center", ds.arr(center, "code_length"))
        n_outside = (dd[("index", "radius")] >= 0.25).sum()
        assert_equal(
            data[("index", "radius")].size + n_outside, dd[("index", "radius")].size
        )

        positions = np.array([data[("index", ax)] for ax in "xyz"])
        centers = (
            np.tile(data.center, data[("index", "x")].shape[0])
            .reshape(data[("index", "x")].shape[0], 3)
            .transpose()
        )
        dist = periodic_dist(
            positions,
            centers,
            ds.domain_right_edge - ds.domain_left_edge,
            ds.periodicity,
        )
        # WARNING: this value has not been externally verified
        assert_array_less(dist, 0.25)


def test_ellipsoid_selector():
    # generate fake data with a number of non-cubical grids
    ds = fake_random_ds(64, nprocs=51)
    assert all(ds.periodicity)

    ellipsoids = [[0.0, 0.0, 0.0], [0.5, 0.5, 0.5], [1.0, 1.0, 1.0], [0.25, 0.75, 0.25]]

    # spherical ellipsoid tests
    ratios = 3 * [0.25]
    for center in ellipsoids:
        data = ds.ellipsoid(
            center, ratios[0], ratios[1], ratios[2], np.array([1.0, 0.0, 0.0]), 0.0
        )
        data.get_data()

        dd = ds.all_data()
        dd.set_field_parameter("center", ds.arr(center, "code_length"))
        n_outside = (dd[("index", "radius")] >= ratios[0]).sum()
        assert_equal(
            data[("index", "radius")].size + n_outside, dd[("index", "radius")].size
        )

        positions = np.array([data[("index", ax)] for ax in "xyz"])
        centers = (
            np.tile(data.center, data.shape[0]).reshape(data.shape[0], 3).transpose()
        )
        dist = periodic_dist(
            positions,
            centers,
            ds.domain_right_edge - ds.domain_left_edge,
            ds.periodicity,
        )
        # WARNING: this value has not been externally verified
        assert_array_less(dist, ratios[0])

    # aligned ellipsoid tests
    ratios = [0.25, 0.1, 0.1]
    for center in ellipsoids:
        data = ds.ellipsoid(
            center, ratios[0], ratios[1], ratios[2], np.array([1.0, 0.0, 0.0]), 0.0
        )

        # hack to compute elliptic distance
        dist2 = np.zeros(data[("index", "ones")].shape[0])
        for i, ax in enumerate(("index", k) for k in "xyz"):
            positions = np.zeros((3, data[("index", "ones")].shape[0]))
            positions[i, :] = data[ax]
            centers = np.zeros((3, data[("index", "ones")].shape[0]))
            centers[i, :] = center[i]
            dist2 += (
                periodic_dist(
                    positions,
                    centers,
                    ds.domain_right_edge - ds.domain_left_edge,
                    ds.periodicity,
                )
                / ratios[i]
            ) ** 2
        # WARNING: this value has not been externally verified
        assert_array_less(dist2, 1.0)


def test_slice_selector():
    # generate fake data with a number of non-cubical grids
    ds = fake_random_ds(64, nprocs=51)
    assert all(ds.periodicity)

    for i, d in enumerate(("index", k) for k in "xyz"):
        for coord in np.arange(0.0, 1.0, 0.1):
            data = ds.slice(i, coord)
            data.get_data()
            v = data[d].to_ndarray()
            assert_equal(data.shape[0], 64**2)
            assert_equal(data[("index", "ones")].shape[0], 64**2)
            assert_array_less(np.abs(v - coord), 1.0 / 128.0 + 1e-6)


def test_cutting_plane_selector():
    # generate fake data with a number of non-cubical grids
    ds = fake_random_ds(64, nprocs=51)
    assert all(ds.periodicity)

    # test cutting plane against orthogonal plane
    for i in range(3):
        norm = np.zeros(3)
        norm[i] = 1.0

        for coord in np.arange(0, 1.0, 0.1):
            center = np.zeros(3)
            center[i] = coord

            data = ds.slice(i, coord)
            data.get_data()
            data2 = ds.cutting(norm, center)
            data2.get_data()

            assert data.shape[0] == data2.shape[0]

            cells1 = np.lexsort(
                (data[("index", "x")], data[("index", "y")], data[("index", "z")])
            )
            cells2 = np.lexsort(
                (data2[("index", "x")], data2[("index", "y")], data2[("index", "z")])
            )
            for d2 in "xyz":
                assert_equal(data[("index", d2)][cells1], data2[("index", d2)][cells2])


# def test_region_selector():
#
# def test_disk_selector():
#
# def test_orthoray_selector():
#
# def test_ray_selector():



import numpy as np
from numpy.testing import assert_raises

from yt import ProjectionPlot, load_amr_grids
from yt.utilities.exceptions import YTIllDefinedAMR, YTIntDomainOverflow


def test_qt_overflow():
    grid_data = []

    grid_dict = {}

    grid_dict["left_edge"] = [-1.0, -1.0, -1.0]
    grid_dict["right_edge"] = [1.0, 1.0, 1.0]
    grid_dict["dimensions"] = [8, 8, 8]
    grid_dict["level"] = 0

    grid_dict["density"] = np.ones((8, 8, 8))

    grid_data.append(grid_dict)

    domain_dimensions = np.array([8, 8, 8])

    spf = load_amr_grids(grid_data, domain_dimensions)

    def make_proj():
        p = ProjectionPlot(spf, "x", [("gas", "density")], center="c", origin="native")
        return p

    assert_raises(YTIntDomainOverflow, make_proj)


def test_refine_by():
    grid_data = []
    ref_by = 4
    lo = 0.0
    hi = 1.0
    fine_grid_width = (hi - lo) / ref_by
    for level in range(2):
        grid_dict = {}

        grid_dict["left_edge"] = [0.0 + 0.5 * fine_grid_width * level] * 3
        grid_dict["right_edge"] = [1.0 - 0.5 * fine_grid_width * level] * 3
        grid_dict["dimensions"] = [8, 8, 8]
        grid_dict["level"] = level

        grid_dict["density"] = np.ones((8, 8, 8))

        grid_data.append(grid_dict)

    domain_dimensions = np.array([8, 8, 8])

    load_amr_grids(grid_data, domain_dimensions, refine_by=ref_by)


def test_validation():
    dims = np.array([4, 2, 4])
    grid_data = [
        {
            "left_edge": [0.0, 0.0, 0.0],
            "right_edge": [1.0, 1.0, 1.0],
            "level": 0,
            "dimensions": dims,
        },
        {
            "left_edge": [0.25, 0.25, 0.25],
            "right_edge": [0.75, 0.75, 0.75],
            "level": 1,
            "dimensions": dims,
        },
    ]
    bbox = np.array([[0, 1], [0, 1], [0, 1]])

    def load_grids():
        load_amr_grids(
            grid_data,
            dims,
            bbox=bbox,
            periodicity=(0, 0, 0),
            length_unit=1.0,
            refine_by=2,
        )

    assert_raises(YTIllDefinedAMR, load_grids)
